# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serializers for sessions."""

from typing import List, Optional

from renku.command.format.tabulate import tabulate
from renku.domain_model.session import Session


def tabular(sessions: List[Session], *, columns: Optional[str] = None):
    """Format sessions with a tabular output."""
    if not columns:
        columns = "id,start_time,status,provider,url"

        if any(s.ssh_enabled for s in sessions):
            columns += ",ssh"

    return tabulate(collection=sessions, columns=columns, columns_mapping=SESSION_COLUMNS)


def log(sessions: List[Session], *, columns: Optional[str] = None):
    """Format sessions in a log like output."""
    from renku.ui.cli.utils.terminal import style_header, style_key

    output = []

    for session in sessions:
        output.append(style_header(f"Session {session.id} ({session.status}, {session.provider})"))
        output.append(style_key("Started: ") + session.start_time.isoformat())
        output.append(style_key("Url: ") + session.url)
        output.append(style_key("Commit: ") + session.commit)
        output.append(style_key("Branch: ") + session.branch)
        output.append(style_key("SSH enabled: ") + ("yes" if session.ssh_enabled else "no"))
        output.append("")
    return "\n".join(output)


SESSION_FORMATS = {"tabular": tabular, "log": log}
"""Valid formatting options."""

SESSION_COLUMNS = {
    "id": ("id", "id"),
    "status": ("status", "status"),
    "url": ("url", "url"),
    "ssh": ("ssh_enabled", "SSH enabled"),
    "start_time": ("start_time", "start_time"),
    "commit": ("commit", "commit"),
    "branch": ("branch", "branch"),
    "provider": ("provider", "provider"),
}



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Renku API utilities."""

from functools import wraps
from typing import TYPE_CHECKING, Optional

from renku.domain_model.project_context import project_context

if TYPE_CHECKING:
    from renku.infrastructure.gateway.activity_gateway import ActivityGateway
    from renku.infrastructure.gateway.plan_gateway import PlanGateway


def ensure_project_context(fn):
    """Check existence of a project context.

    Args:
        fn: The function to wrap.

    Returns:
        The function with the current project injected.
    """

    def get_current_project():
        """Return current project context if any or a new project object.

        Returns:
            The current project context or None.
        """
        from renku.ui.api import Project

        return Project._project_contexts.top if Project._project_contexts.top else None

    @wraps(fn)
    def wrapper(*args, **kwargs):
        from renku.ui.api import Project

        project = get_current_project() or Project()
        return fn(*args, **kwargs, project=project)

    return wrapper


@ensure_project_context
def get_activity_gateway(project) -> Optional["ActivityGateway"]:
    """Return an instance of ActivityGateway when inside a Renku project."""
    from renku.infrastructure.gateway.activity_gateway import ActivityGateway

    try:
        _ = project_context.repository
    except ValueError:
        return None

    return ActivityGateway()


@ensure_project_context
def get_plan_gateway(project) -> Optional["PlanGateway"]:
    """Return an instance of PlanGateway when inside a Renku project."""
    from renku.infrastructure.gateway.plan_gateway import PlanGateway

    try:
        _ = project_context.repository
    except ValueError:
        return None

    return PlanGateway()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Check your system and repository for potential problems."""

import traceback

from pydantic import validate_arguments

from renku.command.command_builder.command import Command
from renku.command.util import ERROR

DOCTOR_INFO = """\
Please note that the diagnosis report is used to help Renku maintainers with
debugging if you file an issue. Use all proposed solutions with maximal care
and if in doubt ask an expert around or file an issue. Thanks!
"""


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def _doctor_check(fix: bool, force: bool):
    """Check your system and repository for potential problems.

    Args:
        fix(bool): Whether to apply fixes or just check.
        force(bool): Whether to force-fix some actions.

    Returns:
        Tuple of whether the project is ok or not and list of problems found.
    """
    from renku.command import checks

    is_ok = True
    fixes_available = False
    problems = []

    for check in checks.__all__:
        try:
            ok, has_fix, problems_ = getattr(checks, check)(fix=fix, force=force)
        except Exception:
            ok = False
            has_fix = False
            tb = "\n\t".join(traceback.format_exc().split("\n"))
            problems_ = f"{ERROR}Exception raised when running {check}\n\t{tb}"

        is_ok &= ok
        fixes_available |= has_fix

        if problems_:
            problems.append(problems_)

    return is_ok, fixes_available, "\n".join(problems)


def doctor_check_command(with_fix):
    """Command to check your system and repository for potential problems.

    Args:
        with_fix: Whether to fix found problems or just check.

    Returns:
        Tuple of whether the project is ok or not and list of problems found.
    """
    if with_fix:
        return (
            Command()
            .command(_doctor_check)
            .lock_project()
            .with_database(write=True)
            .require_migration()
            .require_clean()
            .with_commit()
        )
    else:
        return Command().command(_doctor_check).with_database()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migration models V10."""

from datetime import datetime
from typing import TYPE_CHECKING, List, Optional, cast
from urllib.parse import quote

import persistent

from renku.core.util.datetime8601 import fix_datetime, local_now, parse_date
from renku.core.util.git import get_git_user
from renku.core.util.os import normalize_to_ascii
from renku.domain_model.provenance.agent import Person
from renku.domain_model.provenance.annotation import Annotation
from renku.version import __minimum_project_version__

if TYPE_CHECKING:
    from renku.domain_model.project_context import ProjectRemote
    from renku.infrastructure.repository import Repository


class Project(persistent.Persistent):
    """Represent a project."""

    keywords: List[str] = list()

    # NOTE: the minimum version of renku to needed to work with a project
    # This should be bumped on metadata version changes and when we do not forward-compatible on-the-fly migrations
    minimum_renku_version: str = __minimum_project_version__

    def __init__(
        self,
        *,
        agent_version: Optional[str] = None,
        annotations: Optional[List[Annotation]] = None,
        automated_update: bool = False,
        creator: Person,
        date_created: Optional[datetime] = None,
        description: Optional[str] = None,
        id: Optional[str] = None,
        immutable_template_files: Optional[List[str]] = None,
        name: Optional[str] = None,
        template_id: Optional[str] = None,
        template_metadata: str = "{}",
        template_ref: Optional[str] = None,
        template_source: Optional[str] = None,
        template_version: Optional[str] = None,
        version: Optional[str] = None,
        keywords: Optional[List[str]] = None,
    ):
        from renku.core.migration.migrate import SUPPORTED_PROJECT_VERSION

        version = cast(str, version or SUPPORTED_PROJECT_VERSION)
        date_created = parse_date(date_created) or local_now()

        if id is None:
            namespace, generated_name = Project.get_namespace_and_name(name=name, creator=creator)
            assert generated_name is not None, "Cannot generate Project id with no name"
            id = Project.generate_id(namespace=namespace, name=generated_name)

        self.agent_version: Optional[str] = agent_version
        self.annotations: List[Annotation] = annotations or []
        self.automated_update: bool = automated_update
        self.creator: Person = creator
        self.date_created: datetime = fix_datetime(date_created) or local_now()
        self.description: Optional[str] = description
        self.id: str = id
        self.immutable_template_files: Optional[List[str]] = immutable_template_files
        self.name: Optional[str] = name
        self.template_id: Optional[str] = template_id
        self.template_metadata: str = template_metadata
        self.template_ref: Optional[str] = template_ref
        self.template_source: Optional[str] = template_source
        self.template_version: Optional[str] = template_version
        self.version: str = version
        self.keywords = keywords or []

        # NOTE: We copy this over as class variables don't get saved in the DB
        self.minimum_renku_version = Project.minimum_renku_version

    @staticmethod
    def get_namespace_and_name(
        *,
        remote: Optional["ProjectRemote"] = None,
        repository: Optional["Repository"] = None,
        name: Optional[str] = None,
        namespace: Optional[str] = None,
        creator: Optional[Person] = None,
    ):
        """Return Project's namespace and name from various objects."""
        if remote:
            namespace = namespace or remote.owner
            name = name or remote.name

        if not creator and repository:
            creator = get_git_user(repository=repository)

        if not namespace and creator:
            namespace = creator.email.split("@")[0]

        return namespace, name

    @staticmethod
    def generate_id(namespace: str, name: str):
        """Generate an id for Project."""
        assert namespace, "Cannot generate Project id with no namespace"
        assert name, "Cannot generate Project id with no name"

        namespace = quote(namespace.strip("/"), safe="/")
        slug = normalize_to_ascii(name)

        return f"/projects/{namespace}/{slug}"



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Create an empty Renku project or reinitialize an existing one.

Description
~~~~~~~~~~~

Use existing renku templates, either official or from the community, to
create a new Renku project or Renku-ize an existing one.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.init:init
   :prog: renku init
   :nested: full

Use a different template
~~~~~~~~~~~~~~~~~~~~~~~~

Renku is installed together with a specific set of templates you can select
when you initialize a project. You can check them by typing:

.. code-block:: console

    $ renku template ls

    INDEX ID     DESCRIPTION                     PARAMETERS
    ----- ------ ------------------------------- -----------------------------
    1     python The simplest Python-based [...] description: project des[...]
    2     R      R-based renku project with[...] description: project des[...]

If you know which template you are going to use, you can provide its id using
``--template-id``.

You can use a newer version of the templates or even create your own one and
provide it to the ``init`` command by specifying the target template repository
source ``--template-source`` (both local path and remote url are supported) and
the reference ``--template-ref`` (branch, tag or commit).

You can take inspiration from the
`official Renku template repository
<https://github.com/SwissDataScienceCenter/renku-project-template>`_

.. code-block:: console

    $ renku init --template-ref master --template-source \
    https://github.com/SwissDataScienceCenter/renku-project-template

    Fetching template from
    https://github.com/SwissDataScienceCenter/renku-project-template@master
    ... OK

    NUMBER ID             DESCRIPTION                PARAMETERS
    ----- -------------- -------------------------- ----------------------
    1     python-minimal Basic Python Project:[...] description: proj[...]
    2     R-minimal      Basic R Project: The [...] description: proj[...]

    Please choose a template by typing the number:

Provide parameters
~~~~~~~~~~~~~~~~~~

Some templates require parameters to properly initialize a new project. You
can check them by listing the templates ``renku template ls --verbose``.

To provide parameters, use the ``--parameter`` option and provide each
parameter using ``--parameter "param1"="value1"``.

.. code-block:: console

    $ renku init --template-id python-minimal --parameter \
    "description"="my new shiny project"

    Initializing new Renku repository... OK

If you don't provide the required parameters through the option
``-parameter``, you will be asked to provide them. Empty values are allowed
and passed to the template initialization function.

.. note::

    Project's ``name`` is considered as a special parameter and it's
    automatically added to the list of parameters forwarded to the ``init``
    command.

Provide custom metadata
~~~~~~~~~~~~~~~~~~~~~~~

Custom metadata can be added to the projects knowledge graph by writing
it to a json file and passing that via the `--metadata` option.

.. code-block:: console

    $ echo '{"@id": "https://example.com/id1", \
        "@type": "https://schema.org/Organization", \
        "https://schema.org/legalName": "ETHZ"}' > metadata.json

    $ renku init --template-id python-minimal --parameter \
    "description"="my new shiny project" --metadata metadata.json

    Initializing new Renku repository... OK


Update an existing project
~~~~~~~~~~~~~~~~~~~~~~~~~~

There are situations when the required structure of a Renku project needs
to be recreated or you have an **existing** Git repository for folder that
you wish to turn into a Renku project. In these cases, Renku will warn you
if there are any files that need to be overwritten. ``README.md`` and
``README.rst`` will never be overwritten. ``.gitignore`` will be appended to
to prevent files accidentally getting committed. Files that are not present
in the template will be left untouched by the command.

.. code-block:: console

    $ echo "# Example\nThis is a README." > README.md
    $ echo "FROM python:3.10-alpine" > Dockerfile
    $ renku init

    NUMBER  ID              PARAMETERS
    -------  --------------  ------------
        1  python-minimal  description
        2  R-minimal       description
        3  bioc-minimal    description
        4  julia-minimal   description
        5  minimal
    Please choose a template by typing the number: 1
    The template requires a value for "description": Test Project
    Initializing Git repository...
    Warning: The following files exist in the directory and will be overwritten:
            Dockerfile
    Proceed? [y/N]: y
    Initializing new Renku repository...
    Initializing file .dockerignore ...
    Initializing file .gitignore ...
    Initializing file .gitlab-ci.yml ...
    Initializing file .renku/renku.ini ...
    Initializing file .renkulfsignore ...
    Overwriting file Dockerfile ...
    Initializing file data/.gitkeep ...
    Initializing file environment.yml ...
    Initializing file notebooks/.gitkeep ...
    Initializing file requirements.txt ...
    Project initialized.
    OK

If you initialize in an existing git repository, Renku will create a backup
branch before overwriting any files and will print commands to revert the
changes done and to see what changes were made.

You can also enable the external storage system for output files, if it
was not installed previously.

.. code-block:: console

    $ renku init --external-storage

"""

import json
import os
from pathlib import Path

import click

from renku.command.options import option_external_storage_requested
from renku.core import errors
from renku.ui.cli.utils import color

INVALID_DATA_DIRS = [".", ".renku", ".git"]
"""Paths that cannot be used as data directory name."""


def parse_parameters(ctx, param, value):
    """Parse parameters to dictionary."""
    parameters = {}
    for parameter in value:
        splitted = parameter.split("=", 1)
        if len(splitted) < 2 or len(splitted[0]) < 1:
            raise errors.ParameterError(
                'Parameter format must be --parameter "param1"="value". ', f'--parameter "{parameter}"'
            )
        parameters[splitted[0]] = splitted[1]
    return parameters


def resolve_data_directory(data_dir, path):
    """Check data directory is within the project path."""
    if not data_dir:
        return

    absolute_data_dir = (Path(path) / data_dir).resolve()

    try:
        data_dir = absolute_data_dir.relative_to(path)
    except ValueError:
        raise errors.ParameterError(f"Data directory {data_dir} is not within project {path}")

    if str(data_dir).rstrip(os.path.sep) in INVALID_DATA_DIRS:
        raise errors.ParameterError(f"Cannot use {data_dir} as data directory.")

    return data_dir


@click.command()
@click.argument("path", default=".", type=click.Path(writable=True, file_okay=False, resolve_path=True))
@click.option("-n", "--name", default=None, help="Provide a custom project name.")
@click.option("--description", help="Provide a description for the project.")
@click.option("-k", "--keyword", default=None, multiple=True, type=click.STRING, help="List of keywords.")
@click.option(
    "--datadir",
    default=None,
    type=click.Path(writable=True, file_okay=False),
    help="Data directory within the project",
)
@click.option("-i", "--image", default=None, type=str, help="Path or URL to project's avatar/image.")
@click.option("-t", "--template-id", help="Provide the id of the template to use.")
@click.option("-s", "--template-source", help="Provide the templates repository url or path.")
@click.option(
    "-r", "--template-ref", default=None, help="Specify the reference to checkout on remote template repository."
)
@click.option(
    "-p",
    "--parameter",
    "parameters",
    multiple=True,
    type=click.STRING,
    callback=parse_parameters,
    help=(
        "Provide parameters value. Should be invoked once per parameter. "
        'Please specify the values as follow: --parameter "param1"="value"'
    ),
)
@click.option(
    "-m",
    "--metadata",
    "metadata",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Custom metadata to be associated with the project.",
)
@click.option("-l", "--list-templates", is_flag=True, help="List templates available in the template-source.")
@click.option("-d", "--describe", is_flag=True, help="Show description for templates and parameters")
@click.option("--force", is_flag=True, help="Override target path.")
@click.option("--initial-branch", help="Initial git branch to create.")
@option_external_storage_requested
@click.pass_context
def init(
    ctx,
    external_storage_requested,
    path,
    name,
    description,
    image,
    keyword,
    template_id,
    template_source,
    template_ref,
    parameters,
    metadata,
    list_templates,
    force,
    describe,
    datadir,
    initial_branch,
):
    """Initialize a project in PATH. Default is the current path."""
    from renku.command.init import init_project_command
    from renku.core.constant import FILESYSTEM_ROOT
    from renku.core.image import ImageObjectRequest
    from renku.core.util.git import check_global_git_user_is_configured
    from renku.ui.cli.utils.callback import ClickCallback

    if template_ref and not template_source:
        raise errors.ParameterError("Can't use '--template-ref' without specifying '--template-source'")
    if describe:
        raise errors.ParameterError("'-d/--describe' is deprecated: Use 'renku template show' instead.")
    if list_templates:
        raise errors.ParameterError("'-l/--list-templates' is deprecated: Use 'renku template ls' instead.")

    datadir = resolve_data_directory(datadir, path)

    check_global_git_user_is_configured()

    custom_metadata = None
    if metadata:
        custom_metadata = json.loads(Path(metadata).read_text())

    image_request = ImageObjectRequest(content_url=image, safe_image_paths=[FILESYSTEM_ROOT]) if image else None

    communicator = ClickCallback()
    init_project_command().with_communicator(communicator).build().execute(
        external_storage_requested=external_storage_requested,
        path=path,
        name=name,
        description=description,
        keywords=keyword,
        image_request=image_request,
        template_id=template_id,
        template_source=template_source,
        template_ref=template_ref,
        input_parameters=parameters,
        custom_metadata=custom_metadata,
        force=force,
        data_dir=datadir,
        initial_branch=initial_branch,
        install_mergetool=True,
    )

    click.secho("OK", fg=color.GREEN)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service config serializers."""

from marshmallow import Schema, fields

from renku.ui.service.serializers.common import (
    AsyncSchema,
    GitCommitSHA,
    GitUrlResponseMixin,
    MigrateSchema,
    RemoteRepositorySchema,
    RenkuSyncSchema,
)
from renku.ui.service.serializers.rpc import JsonRPCResponse


class ConfigShowRequest(RemoteRepositorySchema, GitCommitSHA):
    """Request schema for config show."""


class ConfigShowSchema(Schema):
    """Config generic schema."""

    config = fields.Dict(metadata={"description": "Dictionary of configuration items."}, required=True)


class ConfigShowResponse(ConfigShowSchema, GitUrlResponseMixin):
    """Response schema for project config show."""

    default = fields.Dict(metadata={"description": "Dictionary of default configuration items."}, required=True)


class ConfigShowResponseRPC(JsonRPCResponse):
    """RPC response schema for project config show response."""

    result = fields.Nested(ConfigShowResponse)


class ConfigSetRequest(AsyncSchema, ConfigShowSchema, MigrateSchema, RemoteRepositorySchema):
    """Request schema for config set."""


class ConfigSetResponse(ConfigShowSchema, RenkuSyncSchema, GitUrlResponseMixin):
    """Response schema for project config set."""

    default = fields.Dict(metadata={"description": "Dictionary of default configuration items."})


class ConfigSetResponseRPC(JsonRPCResponse):
    """RPC response schema for project config set response."""

    result = fields.Nested(ConfigSetResponse)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Get and set Renku repository or global options."""

from typing import Dict, Optional

from pydantic import validate_arguments

from renku.command.command_builder.command import Command
from renku.core import errors
from renku.core.config import get_config, get_value, remove_value, set_value
from renku.core.constant import CONFIG_LOCAL_PATH
from renku.domain_model.enums import ConfigFilter


def _split_section_and_key(key):
    """Return a tuple with config section and key.

    Args:
        key: The config key.

    Returns:
        Tuple of "renku" and the supplied key.
    """
    parts = key.split(".")
    if len(parts) > 1:
        return f"{parts[0]}", ".".join(parts[1:])
    return "renku", key


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def _update_multiple_config(
    values: Dict[str, Optional[str]], global_only: bool = False, commit_message: Optional[str] = None
):
    """Add, update, or remove multiple configuration values.

    Args:
        values(Dict[str, str]): Dictionary of config key values to update.
        global_only(bool): Whether to only update global config (Default value = False).
        commit_message(Optional[str]): Commit message for change (Default value = None).
    """
    for k, v in values.items():
        if v is not None:
            _update_config(k, value=v, global_only=global_only)
        else:
            _update_config(k, remove=True, global_only=global_only)


def update_multiple_config():
    """Command for updating config."""
    return (
        Command()
        .command(_update_multiple_config)
        .with_database()
        .require_migration()
        .with_commit(commit_if_empty=False, commit_only=[CONFIG_LOCAL_PATH])
    )


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def _update_config(
    key: str,
    *,
    value: Optional[str] = None,
    remove: bool = False,
    global_only: bool = False,
    commit_message: Optional[str] = None,
):
    """Add, update, or remove configuration values.

    Args:
        key(str): Config key.
        value(Optional[str]): Config value (Default value = None).
        remove(bool): Whether to remove values (Default value = False).
        global_only(bool): Whether to only update global config (Default value = False).
        commit_message(Optional[str]): Commit message for change (Default value = None).
    Raises:
        errors.ParameterError: If key wasn't found.
    Returns:
        The modified/removed value.
    """
    section, section_key = _split_section_and_key(key)
    if remove:
        remove_value(section, section_key, global_only=global_only)
    else:
        set_value(section, section_key, value, global_only=global_only)
    return value


def update_config():
    """Command for updating config."""
    return (
        Command()
        .command(_update_config)
        .require_migration()
        .with_commit(commit_if_empty=False, commit_only=[CONFIG_LOCAL_PATH], skip_staging=True)
        .with_database()
    )


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def _read_config(key: Optional[str], config_filter: ConfigFilter = ConfigFilter.ALL, as_string: bool = True):
    """Read configuration.

    Args:
        key: Config key.
        config_filter: Which config files to read from (Default value = ConfigFilter.ALL).
        as_string: Whether to return a string or dictionary (Default value = True).

    Returns:
        String or dictionary containing configuration values.
    """
    if key:
        section, section_key = _split_section_and_key(key)
        value = get_value(section, section_key, config_filter=config_filter)
        if value is None:
            raise errors.ParameterError(f"Key '{key}' not found.")
        return value

    return get_config(config_filter=config_filter, as_string=as_string)


def read_config():
    """Command for updating config."""
    return Command().command(_read_config)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku workflow file format parser."""

from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Generator, List, Optional, Tuple, Type, Union

from renku.core import errors
from renku.core.interface.workflow_file_parser import IWorkflowFileParser
from renku.core.plugin import hookimpl
from renku.core.util.os import get_absolute_path, get_relative_path, safe_read_yaml
from renku.core.workflow.model.workflow_file import BaseParameterType, Input, Output, Parameter, Step, WorkflowFile
from renku.domain_model.project_context import project_context


class RenkuWorkflowFileParser(IWorkflowFileParser):
    """Parser for Renku workflow files."""

    @hookimpl
    def workflow_file_parser(self) -> Tuple[IWorkflowFileParser, str]:
        """Workflow file parser name."""
        return self, "renku"

    @hookimpl
    def parse(self, path: Union[Path, str]) -> WorkflowFile:
        """Parse a given workflow file using the provider.

        Returns:
            WorkflowFile: A ``WorkflowFile`` generated by the provider.
        """
        data = safe_read_yaml(path=path)

        return convert_to_workflow_file(data=data, path=path)


class Attribute:
    """Defines an attribute of a class and include its validation properties."""

    def __init__(
        self,
        name: str,
        *types: Type,
        required: bool = False,
        list_subtype: Optional[Union[Type, Tuple[Type, ...]]] = None,
    ):
        self.name: str = name
        self.types: Tuple[Type, ...] = types
        self.required: bool = required
        self.list_subtype: Optional[Union[Type, Tuple[Type, ...]]] = list_subtype

    def validate(self, value: Any, kind: str):
        """Validate a given value to have correct type."""
        if value is None:
            if not self.required:
                return
            raise errors.ParseError(f"Attribute '{self.name}' in {kind} is required and cannot be None")

        if not isinstance(value, self.types):
            valid_type_names: Union[str, List[str]] = [t.__name__ for t in self.types]
            if len(valid_type_names) == 1:
                valid_type_names = valid_type_names[0]

            raise errors.ParseError(
                f"Invalid type for attribute '{self.name}' in {kind}: Required '{valid_type_names}' but got "
                f"'{type(value).__name__}': {value}"
            )
        elif self.list_subtype:
            subtype = self.list_subtype
            if not isinstance(value, (list, tuple)):
                raise errors.ParseError(
                    f"Expected List[{subtype}] for attribute '{self.name}' in {kind} but got "
                    f"'{type(value).__name__}': {value}"
                )
            for element in value:
                if not isinstance(element, subtype):
                    if isinstance(subtype, tuple):
                        subtype_name = ", ".join(s.__name__ for s in subtype)
                        subtype_name = f"({subtype_name})"
                    else:
                        subtype_name = subtype.__name__
                    raise errors.ParseError(
                        f"Expected '{subtype_name}' for elements of attribute '{self.name}' in {kind} but "
                        f"got '{type(element).__name__}': {element}"
                    )


def convert_to_workflow_file(data: Dict[str, Any], path: Union[Path, str]) -> WorkflowFile:
    """Create an instance of ``WorkflowFile``."""
    if not isinstance(data, dict):
        raise errors.ParseError("Workflow file must be a YAML dictionary")

    validate_attributes(
        kind="workflow file",
        name=None,
        data=data,
        attributes=[
            Attribute("description", str),
            Attribute("keywords", list, list_subtype=(str, int, float)),
            Attribute("name", str, required=True),
            Attribute("steps", list, dict, required=True),
        ],
    )

    if len(data["steps"]) == 0:
        raise errors.ParseError("Workflow file must have at least one step")

    try:
        path = get_absolute_path(path=path)
        relative_path = get_relative_path(path=path, base=project_context.path)
    except errors.ProjectContextError:
        relative_path = str(path)

    # NOTE: We always store the relative path (if available) in the WorkflowFile
    path = relative_path or path
    name: str = data.get("name")  # type: ignore

    return WorkflowFile(
        path=path,
        name=name,
        description=data.get("description"),
        steps=[convert_to_step(data=e, path=path, workflow_file_name=name) for e in get_elements(data["steps"])],
        keywords=data.get("keywords"),
    )


def convert_to_step(data: Union[Dict[str, Any], str], path: Union[Path, str], workflow_file_name: str) -> Step:
    """Create an instance of a workflow file ``Step``."""
    data = convert_element(
        kind="step",
        data=data,
        attributes=[
            Attribute("command", str, required=True),
            Attribute("keywords", list, list_subtype=(str, int, float)),
            Attribute("description", str),
            Attribute("inputs", dict, list),
            Attribute("outputs", dict, list),
            Attribute("parameters", dict, list),
            Attribute("success_codes", list, list_subtype=int),
        ],
        allow_single_value=False,
    )

    if not isinstance(data, dict):
        raise errors.ParseError(
            f"Expected Dict[str, Any] when processing steps, but found '{type(data).__name__}': {data}"
        )
    name = data["name"]

    return Step(
        command=data["command"],
        description=data.get("description"),
        inputs=[convert_to_base_parameter(Input, data=e) for e in get_elements(data.get("inputs", []))],
        keywords=data.get("keywords"),
        name=name,
        outputs=[convert_to_base_parameter(Output, data=e) for e in get_elements(data.get("outputs", []))],
        parameters=[convert_to_base_parameter(Parameter, data=e) for e in get_elements(data.get("parameters", []))],
        path=path,
        success_codes=data.get("success_codes"),
        workflow_file_name=workflow_file_name,
    )


def convert_to_base_parameter(cls: Type[BaseParameterType], data: Union[Dict[str, Any], str]) -> BaseParameterType:
    """Create an instance of a workflow file ``Input``, ``Output``, or ``Parameter``."""
    attributes = [
        Attribute("description", str),
        Attribute("prefix", str),
        Attribute("implicit", bool),
    ]

    if cls == Parameter:
        attributes.append(Attribute("value", str, int, float, bool, required=True))
    else:
        attributes.extend(
            [
                Attribute("path", str, int, float, required=True),
                Attribute("persist", bool),
            ]
        )

    data = convert_element(kind=cls.__name__.lower(), data=data, attributes=attributes)

    if not isinstance(data, dict):
        data = {"value": data} if cls == Parameter else {"path": str(data)}
    elif cls in [Input, Output]:
        # NOTE: Path might be an int/float if user enters something like 123 as the path
        data["path"] = str(data.pop("path"))

    return cls(**data)


def get_elements(data: Union[Dict[str, Any], List[Dict[str, Any]]]) -> Generator[Dict[str, Any], None, None]:
    """Get elements of a list of dict as a dict."""
    if isinstance(data, list):
        for element in data:
            if isinstance(element, list):  # type: ignore[unreachable]
                raise errors.ParseError(f"Data cannot be a list: {element}")
            yield element
    elif not isinstance(data, dict):
        raise errors.ParseError(f"Expected Dict[str, Any] but found '{type(data).__name__}': {data}")
    else:
        for key, value in data.items():
            yield {key: value}


def convert_element(
    kind: str,
    data: Union[Dict[str, Any], str],
    attributes: List[Attribute],
    allow_single_value: bool = True,
) -> Union[Dict[str, Any], str]:
    """Convert possible element definition formats to a dict format."""
    if isinstance(data, dict) and len(data) == 1:  # NOTE: <name>: {<properties>} format
        name, value = list(data.items())[0]

        if not isinstance(value, dict):
            kind = f"{kind} '{name}'" if name else kind
            raise errors.ParseError(
                f"Expected Dict[str, Any] when processing {kind}, but found '{type(value).__name__}': {value}"
            )

        validate_attributes(kind=kind, name=name, data=value, attributes=attributes)
        data = {"name": name}
        data.update(value)
        return data
    else:  # NOTE: List of inputs/outputs/parameters
        if not allow_single_value:
            raise errors.ParseError(f"Expected Dict[str, Any] with one value but got: '{data}'")
        if not isinstance(data, (str, int, float)):
            raise errors.ParseError(f"Expected string, integer, or floating-point type but got '{type(data)}': {data}")
        return str(data)


def validate_attributes(*, kind: str, name: Optional[str], data: Dict[str, Any], attributes: List[Attribute]):
    """Validate that the loaded yaml data is valid."""
    kind = f"{kind} '{name}'" if name else kind

    if attributes:
        valid_attributes = {a.name for a in attributes}
        invalid_attributes = set(data.keys()) - valid_attributes
        if invalid_attributes:
            invalid_attributes_str = ", ".join(sorted(invalid_attributes))
            raise errors.ParseError(f"Invalid attributes for {kind}: {invalid_attributes_str}")

        for attribute in attributes:
            if attribute.name in data:
                attribute.validate(value=data[attribute.name], kind=kind)
            elif attribute.required:
                raise errors.ParseError(f"Required attribute '{attribute.name}' isn't set for {kind}: {data}")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base storage handler."""

import json
import os
import posixpath
import subprocess
from pathlib import Path
from typing import Any, Dict, List, Union

from renku.core import errors
from renku.core.interface.storage import FileHash, IStorage
from renku.domain_model.constant import NO_VALUE


class RCloneStorage(IStorage):
    """External storage implementation that uses RClone."""

    def download(self, uri: str, destination: Union[Path, str]) -> None:
        """Download data from ``uri`` to ``destination``."""
        self.run_command_with_uri("copyto", uri, destination)

    def exists(self, uri: str) -> bool:
        """Checks if a remote storage URI exists."""
        try:
            self.run_command_with_uri("lsf", uri=uri, max_depth=1)
        except errors.StorageObjectNotFound:
            return False
        else:
            return True

    def get_hashes(self, uri: str, hash_type: str = "md5") -> List[FileHash]:
        """Download hashes with rclone and parse them.

        Returns a tuple containing a list of parsed hashes.

        Arguments:
            uri(str): Provider uri.
            hash_type(str): Type of hash to get from rclone (Default value = ``md5``).

        Example:
            hashes_raw json::

                [
                    {
                        "Path":"resources/hg19.window.masker.bed.gz.tbi","Name":"hg19.window.masker.bed.gz.tbi",
                        "Size":578288,"MimeType":"application/x-gzip","ModTime":"2022-02-07T18:45:52.000000000Z",
                        "IsDir":false,"Hashes":{"md5":"e93ac5364e7799bbd866628d66c7b773"},"Tier":"STANDARD"
                    }
                ]
        """
        hashes = self.list_files(
            uri=uri, hash=True, recursive=True, files_only=True, no_modtime=True, no_mimetype=True, hash_type=hash_type
        )
        output = []
        is_directory = self.is_directory(uri)
        for hash in hashes:
            path: str = hash["Path"]
            full_uri = posixpath.join(uri, path.strip("/")) if is_directory else uri
            hash_content = hash.get("Hashes", {}).get(hash_type)
            file = FileHash(uri=full_uri, path=hash["Path"], size=hash.get("Size"), hash=hash_content)
            output.append(file)

        return output

    def is_directory(self, uri: str) -> bool:
        """Return True if URI points to a directory.

        NOTE: This returns True for non-existing paths on bucket-based backends like S3 since listing non-existing paths
        won't fail and there is no way to distinguish between empty directories and non-existing paths.
        """
        uri = uri.rstrip("/")
        try:
            # NOTE: Listing with a trailing slash works for directories
            files = self.list_files(uri=f"{uri}/")
        except errors.StorageObjectNotFound:
            return False
        else:
            # NOTE: Listing with trailing slash won't fail for non-existing directories on S3 and similar backends
            # NOTE: Listing a file returns exactly one entry in the list; also, if a single entry is a directory then
            # its parent is a directory too.
            if len(files) != 1 or files[0]["IsDir"]:
                return True
            pathname = files[0]["Path"]
            if pathname != posixpath.basename(uri):
                return True
            # NOTE: The only remaining possibility is a directory with a single file with the same name (e.g. data/data:
            # Listing data/ and data/data returns the same results)
            try:
                files = self.list_files(uri=f"{uri}/{pathname}")
            except (errors.StorageObjectNotFound, errors.ParameterError):
                return False
            else:
                return True if len(files) == 1 else False

    def mount(self, path: Union[Path, str]) -> None:
        """Mount the provider's URI to the given path."""
        self.run_command_with_uri("mount", self.provider.uri, str(path), daemon=True, read_only=True, no_modtime=True)

    def get_configurations(self) -> Dict[str, str]:
        """Get required configurations for rclone to access the storage."""
        configurations = {}
        for name, value in self.credentials.items():
            if value is not NO_VALUE:
                name = get_rclone_env_var_name(self.storage_scheme, name)
                configurations[name] = value

        for name, value in self._provider_configuration.items():
            name = get_rclone_env_var_name(self.storage_scheme, name)
            configurations[name] = value

        return configurations

    def list_files(self, uri: str, *args, **kwargs) -> List[Dict[str, Any]]:
        """List a URI and return results in JSON format."""
        hashes_raw = self.run_command_with_uri("lsjson", uri, *args, **kwargs)
        try:
            hashes = json.loads(hashes_raw)
        except json.JSONDecodeError as e:
            raise errors.RCloneException(f"Cannot parse command output: {e}")
        if not hashes:
            raise errors.ParameterError(f"Cannot find URI: {uri}")

        return hashes

    def run_command_with_uri(self, command: str, uri: str, *args, **kwargs) -> Any:
        """Run a RClone command by converting a given URI."""
        uri = self.provider.convert_to_storage_uri(uri)

        return self.run_command(command, uri, *args, **kwargs)

    def run_command(self, command: str, *args, **kwargs) -> Any:
        """Run a RClone command with storage-specific configuration."""
        return run_rclone_command(command, *args, **kwargs, env=self.get_configurations())

    def upload(self, source: Union[Path, str], uri: str) -> None:
        """Upload data from ``source`` to ``uri``."""
        uri = self.provider.convert_to_storage_uri(uri)

        self.run_command("copyto", source, uri)


def run_rclone_command(command: str, *args: Any, env=None, **kwargs) -> str:
    """Execute an RClone command."""
    os_env = os.environ.copy()
    if env:
        os_env.update(env)

    full_command = ("rclone", "--config", os.devnull, command, *transform_kwargs(**kwargs), *transform_args(*args))
    try:
        result = subprocess.run(full_command, text=True, capture_output=True, env=os_env)
    except FileNotFoundError:
        raise errors.RCloneException("RClone is not installed. See https://rclone.org/install/")

    # See https://rclone.org/docs/#list-of-exit-codes for rclone exit codes
    if result.returncode == 0:
        return result.stdout

    all_outputs = result.stdout + result.stderr
    if (
        result.returncode in (3, 4)
        or (result.returncode == 1 and "failed to read directory entry" in all_outputs)
        or (result.returncode == 1 and "no Host in request URL" in all_outputs)
    ):
        raise errors.StorageObjectNotFound(all_outputs)
    elif (
        "AccessDenied" in all_outputs
        or "no authentication method configured" in all_outputs
        or "Need account+key" in all_outputs
    ):
        raise errors.AuthenticationError(f"Authentication failed when accessing the cloud storage: {all_outputs}")
    else:
        raise errors.RCloneException(f"Remote storage operation failed: {all_outputs}")


def transform_args(*args) -> List[str]:
    """Transforms args to command line args."""
    return [str(a) for a in args]


def transform_kwargs(**kwargs) -> List[str]:
    """Transforms kwargs to command line args."""

    def transform_kwarg(key: str, value: Any) -> List[str]:
        if value in (False, None):
            return []
        else:
            name = f"-{key}" if len(key) == 1 else f"--{key.replace('_', '-')}"
            return [name] if value is True else [name, f"{value}"]

    all_args = []
    for key, value in kwargs.items():
        args = transform_kwarg(key, value)
        all_args.extend(args)

    return all_args


def get_rclone_env_var_name(provider_name, name) -> str:
    """Get name of an RClone env var config."""
    # See https://rclone.org/docs/#config-file
    name = name.replace(" ", "_").replace("-", "_")
    return f"RCLONE_CONFIG_{provider_name}_{name}".upper()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Project JSON-LD schema."""

from marshmallow import EXCLUDE, pre_dump

from renku.command.schema.agent import PersonSchema
from renku.command.schema.annotation import AnnotationSchema
from renku.command.schema.calamus import DateTimeList, JsonLDSchema, Nested, StringList, fields, oa, prov, renku, schema
from renku.command.schema.image import ImageObjectSchema
from renku.domain_model.project import Project


class ProjectSchema(JsonLDSchema):
    """Project Schema."""

    class Meta:
        """Meta class."""

        rdf_type = [schema.Project, prov.Location]
        model = Project
        unknown = EXCLUDE

    agent_version = StringList(schema.agent, load_default="pre-0.11.0")
    annotations = Nested(oa.hasTarget, AnnotationSchema, reverse=True, many=True)
    creator = Nested(schema.creator, PersonSchema, load_default=None)
    date_created = DateTimeList(schema.dateCreated, load_default=None, format="iso", extra_formats=("%Y-%m-%d",))
    description = fields.String(schema.description, load_default=None)
    id = fields.Id(load_default=None)
    image = fields.Nested(schema.image, ImageObjectSchema, load_default=None)
    immutable_template_files = fields.List(
        renku.immutableTemplateFiles,
        fields.String(),
        load_default=list(),
        attribute="template_metadata.immutable_template_files",
    )
    name = fields.String(schema.name, load_default=None)
    template_id = fields.String(renku.templateId, load_default=None, attribute="template_metadata.template_id")
    template_metadata = fields.String(renku.templateMetadata, load_default=None, attribute="template_metadata.metadata")
    template_ref = fields.String(renku.templateReference, load_default=None, attribute="template_metadata.template_ref")
    template_source = fields.String(
        renku.templateSource, load_default=None, attribute="template_metadata.template_source"
    )
    template_version = fields.String(
        renku.templateVersion, load_default=None, attribute="template_metadata.template_version"
    )
    version = StringList(schema.schemaVersion, load_default="1")
    keywords = fields.List(schema.keywords, fields.String(), load_default=None)

    @pre_dump(pass_many=True)
    def removes_ms(self, objs, many, **kwargs):
        """Remove milliseconds from datetimes.

        Note: since DateField uses `strftime` as format, which only supports timezone info without a colon
        e.g. `+0100` instead of `+01:00`, we have to deal with milliseconds manually instead of using a format string.
        """

        def _replace_times(obj):
            obj.date_created = obj.date_created.replace(microsecond=0)

        if many:
            for obj in objs:
                _replace_times(obj)
            return objs

        _replace_times(objs)
        return objs



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes to represent inputs/outputs/parameters in a Plan."""

import copy
import shlex
import urllib
from abc import abstractmethod
from pathlib import Path, PurePosixPath
from typing import Any, Iterator, List, Optional
from uuid import uuid4

from renku.core import errors
from renku.core.errors import ParameterError
from renku.core.util.urls import get_slug

RANDOM_ID_LENGTH = 4
DIRECTORY_MIME_TYPE = "inode/directory"


def _validate_mime_type(encoding_format: List[str]):
    """Validates MIME-types."""
    if encoding_format and DIRECTORY_MIME_TYPE in encoding_format and len(encoding_format) > 1:
        raise ParameterError(f"Directory MIME-type '{DIRECTORY_MIME_TYPE}'.")


class MappedIOStream:
    """Represents an IO stream (stdin, stdout, stderr)."""

    STREAMS = ["stdin", "stdout", "stderr"]

    def __init__(self, *, id: Optional[str] = None, stream_type: str):
        assert stream_type in MappedIOStream.STREAMS

        self.id: str = id or MappedIOStream.generate_id(stream_type)
        self.stream_type = stream_type

    @classmethod
    def from_str(cls, stream_type: str) -> "MappedIOStream":
        """Create an instance from the given stream type string."""
        if stream_type not in MappedIOStream.STREAMS:
            streams = ", ".join(MappedIOStream.STREAMS)
            raise errors.ParameterError(f"'{stream_type}' must be one of {streams}")

        return cls(stream_type=stream_type)

    @staticmethod
    def generate_id(stream_type: str) -> str:
        """Generate an id for parameters."""
        return f"/iostreams/{stream_type}"


class CommandParameterBase:
    """Represents a parameter for a Plan."""

    # NOTE: This attribute is only used by workflow-file machinery to check if plans are the same or not. We need it,
    # because names are generated randomly when not set by users which make the comparison return incorrect result.
    name_set_by_user: bool = False

    def __init__(
        self,
        *,
        default_value: Any,
        description: Optional[str],
        id: str,
        name: Optional[str],
        name_set_by_user: bool = False,
        position: Optional[int] = None,
        prefix: Optional[str] = None,
        derived_from: Optional[str] = None,
        postfix: Optional[str] = None,
    ):
        self.default_value: Any = default_value
        self.description: Optional[str] = description
        self.id: str = id
        self.position: Optional[int] = position
        self.prefix: Optional[str] = prefix
        self._v_actual_value = None
        self._v_actual_value_set: bool = False
        self.derived_from: Optional[str] = derived_from
        # NOTE: ``postfix`` is used only to generate a nicer ``id`` for a parameter. Its value isn't used anywhere else.
        self.postfix: Optional[str] = postfix
        self.name_set_by_user: bool = name_set_by_user

        if name is not None:
            self.name: str = name
        else:
            self.name = self._get_default_name()

    @staticmethod
    def _generate_id(
        plan_id: str,
        parameter_type: str,
        position: Optional[int],
        postfix: Optional[str] = None,
        name: Optional[str] = None,
    ) -> str:
        """Generate an id for parameters."""
        # /plans/723fd784-9347-4081-84de-a6dbb067545b/inputs/1
        # /plans/723fd784-9347-4081-84de-a6dbb067545b/inputs/stdin
        # /plans/723fd784-9347-4081-84de-a6dbb067545b/inputs/data.csv
        # /plans/723fd784-9347-4081-84de-a6dbb067545b/inputs/dda5fc0f00984917be46dc12f5f7b675
        id = uuid4().hex

        if postfix:
            id = urllib.parse.quote(postfix)
        elif position is not None:
            id = str(position)
        elif name:
            id = name

        return f"{plan_id}/{parameter_type}/{id}"

    def __repr__(self):
        return f"<{self.__class__.__name__} '{self.actual_value}'>"

    @property
    def role(self) -> str:
        """Return a unique role for this parameter within its Plan."""
        assert self.id, "Id is not set"
        return PurePosixPath(self.id).name

    @staticmethod
    def _get_equality_attributes() -> List[str]:
        """Return a list of attributes values that determine if instances are equal."""
        # NOTE: We treat name differently
        return ["description", "default_value", "prefix", "position"]

    def is_equal_to(self, other) -> bool:
        """Return if attributes that cause a change in the parameter, are the same."""
        if self.name_set_by_user != other.name_set_by_user or (self.name_set_by_user and self.name != other.name):
            return False
        return all(getattr(self, a) == getattr(other, a) for a in self._get_equality_attributes())

    def to_argv(self, quote_string: bool = True) -> List[Any]:
        """String representation (sames as cmd argument)."""
        value = self.actual_value

        if isinstance(value, str) and quote_string and " " in value:
            value = shlex.quote(value)

        if self.prefix:
            if self.prefix.endswith(" "):
                return [self.prefix[:-1], str(value)]
            return [f"{self.prefix}{value}"]

        return [str(value)]

    def _update_from(self, other: "CommandParameterBase"):
        """Update this parameter with values from another parameter, if applicable."""
        if other.prefix is not None:
            self.prefix = other.prefix
        if other.position is not None:
            self.position = other.position
        if other.description:
            self.description = other.description
        if other.name_set_by_user:
            self.name_set_by_user = other.name_set_by_user

    @property
    def actual_value(self):
        """Get the actual value to be used for execution."""
        if getattr(self, "_v_actual_value_set", False):
            return self._v_actual_value

        return self.default_value

    @actual_value.setter
    def actual_value(self, value):
        """Set the actual value to be used for execution."""
        self._v_actual_value = value
        self._v_actual_value_set = True

    @property
    def actual_value_set(self):
        """Whether the actual_value on this parameter has been set at least once."""
        return getattr(self, "_v_actual_value_set", False)

    def _generate_name(self, base) -> str:
        return generate_parameter_name(parameter=self, kind=base)

    def _get_default_name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def derive(self, plan_id: str) -> "CommandParameterBase":
        """Create a new command parameter from self."""
        raise NotImplementedError


class CommandParameter(CommandParameterBase):
    """An argument to a command that is neither input nor output."""

    def __init__(
        self,
        *,
        default_value: Any = None,
        description: Optional[str] = None,
        id: str,
        name: Optional[str] = None,
        name_set_by_user: bool = False,
        position: Optional[int] = None,
        prefix: Optional[str] = None,
        derived_from: Optional[str] = None,
        postfix: Optional[str] = None,
    ):
        super().__init__(
            default_value=default_value,
            description=description,
            id=id,
            name=name,
            name_set_by_user=name_set_by_user,
            position=position,
            prefix=prefix,
            derived_from=derived_from,
            postfix=postfix,
        )

    @staticmethod
    def generate_id(
        plan_id: str, position: Optional[int] = None, postfix: Optional[str] = None, name: Optional[str] = None
    ) -> str:
        """Generate an id for CommandParameter."""
        return CommandParameterBase._generate_id(
            plan_id, parameter_type="parameters", position=position, postfix=postfix, name=name
        )

    def __repr__(self):
        return (
            f"<Parameter '{self.name}': {self.actual_value} (default: {self.default_value}, prefix: {self.prefix}, "
            f"position: {self.position})>"
        )

    def _get_default_name(self) -> str:
        return self._generate_name(base="parameter")

    def derive(self, plan_id: str) -> "CommandParameter":
        """Create a new ``CommandParameter`` that is derived from self."""
        parameter = copy.copy(self)
        parameter.id = CommandParameter.generate_id(plan_id=plan_id, position=self.position, postfix=self.postfix)
        return parameter

    def update_from(self, other: "CommandParameter"):
        """Update this output with values from another output, if applicable."""
        super()._update_from(other)


class CommandInput(CommandParameterBase):
    """An input to a command."""

    def __init__(
        self,
        *,
        default_value: Optional[Any] = None,
        description: Optional[str] = None,
        id: str,
        mapped_to: Optional[MappedIOStream] = None,
        name: Optional[str] = None,
        name_set_by_user: bool = False,
        position: Optional[int] = None,
        prefix: Optional[str] = None,
        encoding_format: Optional[List[str]] = None,
        derived_from: Optional[str] = None,
        postfix: Optional[str] = None,
    ):
        assert isinstance(default_value, (Path, str)), f"Invalid value type for CommandOutput: {type(default_value)}"

        super().__init__(
            default_value=str(default_value),
            description=description,
            id=id,
            name=name,
            name_set_by_user=name_set_by_user,
            position=position,
            prefix=prefix,
            derived_from=derived_from,
            postfix=postfix,
        )
        self.mapped_to: Optional[MappedIOStream] = mapped_to

        if encoding_format is not None:
            _validate_mime_type(encoding_format)
        self.encoding_format: Optional[List[str]] = encoding_format

    @staticmethod
    def generate_id(
        plan_id: str, position: Optional[int] = None, postfix: Optional[str] = None, name: Optional[str] = None
    ) -> str:
        """Generate an id for CommandInput."""
        return CommandParameterBase._generate_id(
            plan_id, parameter_type="inputs", position=position, postfix=postfix, name=name
        )

    def to_stream_representation(self) -> str:
        """Input stream representation."""
        return f"< {shlex.quote(self.actual_value)}" if self.mapped_to else ""

    def _get_default_name(self) -> str:
        return self._generate_name(base="input")

    def is_equal_to(self, other) -> bool:
        """Return if attributes that cause a change in the parameter, are the same."""
        if self.mapped_to:
            if not other.mapped_to or self.mapped_to.stream_type != other.mapped_to.stream_type:
                return False
        elif other.mapped_to:
            return False

        return super().is_equal_to(other)

    @staticmethod
    def _get_equality_attributes() -> List[str]:
        """Return a list of attributes values that determine if instances are equal."""
        return CommandParameterBase._get_equality_attributes() + ["encoding_format"]

    def derive(self, plan_id: str) -> "CommandInput":
        """Create a new ``CommandInput`` that is derived from self."""
        parameter = copy.copy(self)
        parameter.id = CommandInput.generate_id(plan_id=plan_id, position=self.position, postfix=self.postfix)
        return parameter

    def update_from(self, other: "CommandInput"):
        """Update this input with values from another input, if applicable."""
        super()._update_from(other)

        if other.encoding_format:
            self.encoding_format = other.encoding_format

        if other.mapped_to is not None:
            self.mapped_to = other.mapped_to


class HiddenInput(CommandInput):
    """An input to a command that is added by Renku and should be hidden from users."""


class CommandOutput(CommandParameterBase):
    """An output from a command."""

    def __init__(
        self,
        *,
        create_folder: bool = False,
        default_value: Any = None,
        description: Optional[str] = None,
        id: str,
        mapped_to: Optional[MappedIOStream] = None,
        name: Optional[str] = None,
        name_set_by_user: bool = False,
        position: Optional[int] = None,
        prefix: Optional[str] = None,
        encoding_format: Optional[List[str]] = None,
        derived_from: Optional[str] = None,
        postfix: Optional[str] = None,
    ):
        assert isinstance(default_value, (Path, str)), f"Invalid value type for CommandOutput: {type(default_value)}"

        super().__init__(
            default_value=str(default_value),
            description=description,
            id=id,
            name=name,
            name_set_by_user=name_set_by_user,
            position=position,
            prefix=prefix,
            derived_from=derived_from,
            postfix=postfix,
        )
        self.create_folder: bool = create_folder
        self.mapped_to: Optional[MappedIOStream] = mapped_to

        if encoding_format is not None:
            _validate_mime_type(encoding_format)
        self.encoding_format: Optional[List[str]] = encoding_format

    @staticmethod
    def generate_id(
        plan_id: str, position: Optional[int] = None, postfix: Optional[str] = None, name: Optional[str] = None
    ) -> str:
        """Generate an id for CommandOutput."""
        return CommandParameterBase._generate_id(
            plan_id, parameter_type="outputs", position=position, postfix=postfix, name=name
        )

    def to_stream_representation(self) -> str:
        """Input stream representation."""
        if not self.mapped_to:
            return ""

        value = shlex.quote(self.actual_value)
        return f"> {value}" if self.mapped_to.stream_type == "stdout" else f" 2> {value}"

    def _get_default_name(self) -> str:
        return self._generate_name(base="output")

    def is_equal_to(self, other) -> bool:
        """Return if attributes that cause a change in the parameter, are the same."""
        if self.mapped_to:
            if not other.mapped_to or self.mapped_to.stream_type != other.mapped_to.stream_type:
                return False
        elif other.mapped_to:
            return False

        return super().is_equal_to(other)

    def update_from(self, other: "CommandOutput"):
        """Update this output with values from another output, if applicable."""
        super()._update_from(other)

        if other.encoding_format:
            self.encoding_format = other.encoding_format

        if other.mapped_to is not None:
            self.mapped_to = other.mapped_to

        if other.create_folder:
            self.create_folder = other.create_folder

    @staticmethod
    def _get_equality_attributes() -> List[str]:
        """Return a list of attributes values that determine if instances are equal."""
        # NOTE: Don't include ``create_folder`` in comparison since its value is state-dependent
        return CommandParameterBase._get_equality_attributes() + ["encoding_format"]

    def derive(self, plan_id: str) -> "CommandOutput":
        """Create a new ``CommandOutput`` that is derived from self."""
        parameter = copy.copy(self)
        parameter.id = CommandOutput.generate_id(plan_id=plan_id, position=self.position, postfix=self.postfix)
        return parameter


class ParameterMapping(CommandParameterBase):
    """A mapping of child parameter(s) to a parent CompositePlan."""

    def __init__(
        self,
        *,
        default_value: Any,
        description: Optional[str] = None,
        id: str,
        name: Optional[str] = None,
        name_set_by_user: bool = False,
        mapped_parameters: List[CommandParameterBase],
        **kwargs,
    ):
        super().__init__(
            default_value=default_value,
            description=description,
            id=id,
            name=name,
            name_set_by_user=name_set_by_user,
            **kwargs,
        )

        self.mapped_parameters: List[CommandParameterBase] = mapped_parameters

    @staticmethod
    def generate_id(
        plan_id: str, position: Optional[int] = None, postfix: Optional[str] = None, name: Optional[str] = None
    ) -> str:
        """Generate an id for ParameterMapping."""
        return CommandParameterBase._generate_id(
            plan_id, parameter_type="mappings", position=position, postfix=postfix, name=name
        )

    def to_stream_representation(self) -> str:
        """Input stream representation."""
        return ""

    def _get_default_name(self) -> str:
        return self._generate_name(base="mapping")

    @CommandParameterBase.actual_value.setter  # type: ignore
    def actual_value(self, value):
        """Set the actual value to be used for execution."""
        self._v_actual_value = value
        self._v_actual_value_set = True

        for mapped_to in self.mapped_parameters:
            if not mapped_to.actual_value_set:
                mapped_to.actual_value = value

    @property
    def leaf_parameters(self) -> Iterator[CommandParameterBase]:
        """Return leaf (non-Mapping) parameters contained by this Mapping."""
        for mapped_to in self.mapped_parameters:
            if isinstance(mapped_to, ParameterMapping):
                yield from mapped_to.leaf_parameters
            else:
                yield mapped_to

    def derive(self, plan_id: str) -> "ParameterMapping":
        """Create a new ``CommandParameter`` that is derived from self."""
        parameter = copy.copy(self)
        parameter.id = ParameterMapping.generate_id(plan_id=plan_id, position=self.position, postfix=self.postfix)
        return parameter


class ParameterLink:
    """A link between a source and one or more sink parameters."""

    def __init__(
        self,
        source: CommandParameterBase,
        sinks: List[CommandParameterBase],
        id: str,
    ):
        self.source = source
        self.sinks = sinks
        self.id: str = id

    def apply(self):
        """Apply source value to sinks."""
        for s in self.sinks:
            s.actual_value = self.source.actual_value

    @staticmethod
    def generate_id(plan_id: str) -> str:
        """Generate an id for parameters."""
        # /plans/723fd784-9347-4081-84de-a6dbb067545b/links/dda5fcbf-0098-4917-be46-dc12f5f7b675
        return f"{plan_id}/links/{uuid4()}"


def generate_parameter_name(parameter, kind) -> str:
    """Generate a name for an input, output, or parameter."""
    name = get_slug(parameter.prefix.strip(" -="), invalid_chars=["."]) if parameter.prefix else kind
    position = parameter.position if parameter.position is not None else uuid4().hex[:RANDOM_ID_LENGTH]
    return f"{name}-{position}"



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logging in to a Renku deployment."""

import os
import time
import urllib
import webbrowser
from typing import TYPE_CHECKING, Optional, cast

from pydantic import validate_arguments

from renku.core import errors
from renku.core.config import get_value, remove_value, set_value
from renku.core.util import communication
from renku.core.util.git import (
    RENKU_BACKUP_PREFIX,
    create_backup_remote,
    get_remote,
    get_renku_repo_url,
    set_git_credential_helper,
)
from renku.core.util.urls import parse_authentication_endpoint
from renku.domain_model.enums import ConfigFilter
from renku.domain_model.project_context import project_context

if TYPE_CHECKING:
    from renku.infrastructure.repository import Repository


CONFIG_SECTION = "http"
KEYCLOAK_REALM = "Renku"
CLIENT_ID = "renku-cli"


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def login(endpoint: Optional[str], git_login: bool, yes: bool):
    """Log into a Renku deployment."""
    from renku.core.util import requests

    try:
        repository = project_context.repository
    except ValueError:
        repository = None

    parsed_endpoint = _parse_endpoint(endpoint)

    remote_name, remote_url = None, None
    if git_login:
        if repository is None:
            communication.warn("Cannot log in to git outside a project.")
        else:
            remote = get_remote(repository)
            if remote:
                remote_name, remote_url = remote.name, remote.url

            if remote_name and remote_url:
                show_login_warning = get_value("renku", "show_login_warning")
                if not yes and (show_login_warning is None or show_login_warning.lower() == "true"):
                    message = (
                        "Remote URL will be changed. Do you want to continue "
                        "(to disable this warning, pass '--yes' or run 'renku config set show_login_warning False')?"
                    )
                    communication.confirm(message, abort=True, warning=True)
            else:
                raise errors.ParameterError("Cannot find a unique remote URL for project.")

    auth_server_url = _get_url(
        parsed_endpoint, path=f"auth/realms/{KEYCLOAK_REALM}/protocol/openid-connect/auth/device"
    )

    try:
        response = requests.post(auth_server_url, data={"client_id": CLIENT_ID})
    except errors.RequestError as e:
        raise errors.RequestError(f"Cannot connect to authorization server at {auth_server_url}.") from e

    requests.check_response(response=response)
    data = response.json()

    verification_uri = data.get("verification_uri")
    user_code = data.get("user_code")
    verification_uri_complete = f"{verification_uri}?user_code={user_code}"

    communication.echo(
        f"Please grant access to '{CLIENT_ID}' in your browser.\n"
        f"If a browser window does not open automatically, go to {verification_uri_complete}"
    )

    webbrowser.open_new_tab(verification_uri_complete)

    polling_interval = min(data.get("interval", 5), 5)
    token_url = _get_url(parsed_endpoint, path=f"auth/realms/{KEYCLOAK_REALM}/protocol/openid-connect/token")
    device_code = data.get("device_code")

    while True:
        time.sleep(polling_interval)

        response = requests.post(
            token_url,
            data={
                "device_code": device_code,
                "client_id": CLIENT_ID,
                "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            },
        )
        status_code = response.status_code
        if status_code == 200:
            break
        elif status_code == 400:
            error = response.json().get("error")

            if error == "authorization_pending":
                continue
            elif error == "slow_down":
                polling_interval += 1
            elif error == "access_denied":
                raise errors.AuthenticationError("Access denied")
            elif error == "expired_token":
                raise errors.AuthenticationError("Session expired, try again")
            else:
                raise errors.AuthenticationError(f"Invalid error message from server: {response.json()}")
        else:
            raise errors.AuthenticationError(f"Invalid status code from server: {status_code} - {response.content}")

    access_token = response.json().get("access_token")
    store_token(parsed_endpoint.netloc, access_token)

    if git_login and repository:
        set_git_credential_helper(repository=cast("Repository", repository), hostname=parsed_endpoint.netloc)
        backup_remote_name, backup_exists, remote = create_backup_remote(
            repository=repository, remote_name=remote_name, url=remote_url  # type:ignore
        )
        if backup_exists:
            communication.echo(f"Backup remote '{backup_remote_name}' already exists.")

        if not remote and not backup_exists:
            communication.error(f"Cannot create backup remote '{backup_remote_name}' for '{remote_url}'")
        else:
            _set_renku_url_for_remote(
                repository=cast("Repository", repository),
                remote_name=remote_name,  # type:ignore
                remote_url=remote_url,  # type:ignore
                hostname=parsed_endpoint.netloc,
            )


def _parse_endpoint(endpoint, use_remote=False):
    parsed_endpoint = parse_authentication_endpoint(endpoint=endpoint, use_remote=use_remote)
    if not parsed_endpoint:
        raise errors.ParameterError("Parameter 'endpoint' is missing.")

    return parsed_endpoint


def _get_url(parsed_endpoint, path, **query_args) -> str:
    query = urllib.parse.urlencode(query_args)
    return parsed_endpoint._replace(path=path, query=query).geturl()


def store_token(hostname: str, access_token: str):
    """Store access token for ``hostname``."""
    set_value(section=CONFIG_SECTION, key=hostname, value=access_token, global_only=True)
    os.chmod(project_context.global_config_path, 0o600)


def _set_renku_url_for_remote(repository: "Repository", remote_name: str, remote_url: str, hostname: str):
    """Set renku repository URL for ``remote_name``.

    Args:
        repository("Repository"): Current ``Repository``.
        remote_name(str): Name of the remote.
        remote_url(str): Url of the remote.
        hostname(str): Hostname.
    Raises:
        errors.GitCommandError: If remote doesn't exist.
    """
    new_remote_url = get_renku_repo_url(remote_url, deployment_hostname=hostname)

    try:
        repository.remotes[remote_name].set_url(url=new_remote_url)
    except errors.GitCommandError as e:
        raise errors.GitError(f"Cannot change remote url for '{remote_name}' to '{new_remote_url}'") from e


def read_renku_token(endpoint: str, get_endpoint_from_remote=False) -> str:
    """Read renku token from renku config file.

    Args:
        endpoint(str):  Endpoint to get token for.
    Keywords:
        get_endpoint_from_remote: if no endpoint is specified, use the repository remote to infer one

    Returns:
        Token for endpoint.
    """
    try:
        parsed_endpoint = _parse_endpoint(endpoint, use_remote=get_endpoint_from_remote)
    except errors.ParameterError:
        return ""
    if not parsed_endpoint:
        return ""

    return _read_renku_token_for_hostname(hostname=parsed_endpoint.netloc)


def _read_renku_token_for_hostname(hostname):
    return get_value(section=CONFIG_SECTION, key=hostname, config_filter=ConfigFilter.GLOBAL_ONLY)


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def logout(endpoint: Optional[str]):
    """Log out from one or all Renku deployments."""
    if endpoint:
        parsed_endpoint = parse_authentication_endpoint(endpoint=endpoint)
        key = parsed_endpoint.netloc
    else:
        key = "*"

    try:
        repository = project_context.repository
    except ValueError:
        repository = None

    remove_value(section=CONFIG_SECTION, key=key, global_only=True)
    _remove_git_credential_helper(repository=repository)
    _restore_git_remote(repository=repository)


def _remove_git_credential_helper(repository):
    if not repository:  # Outside a renku project
        return

    with repository.get_configuration(writable=True) as config:
        try:
            config.remove_value("credential", "helper")
        except errors.GitError:  # NOTE: If already logged out, an exception is raised
            pass


def _restore_git_remote(repository):
    if not repository:  # Outside a renku project
        return

    backup_remotes = [r.name for r in repository.remotes if r.name.startswith(RENKU_BACKUP_PREFIX)]
    for backup_remote in backup_remotes:
        remote_name = backup_remote.replace(f"{RENKU_BACKUP_PREFIX}-", "")
        remote_url = repository.remotes[backup_remote].url

        try:
            repository.remotes[remote_name].set_url(remote_url)
        except errors.GitCommandError:
            raise errors.GitError(f"Cannot restore remote url for '{remote_name}' to {remote_url}")

        try:
            repository.remotes.remove(backup_remote)
        except errors.GitCommandError:
            communication.error(f"Cannot delete backup remote '{backup_remote}'")


def has_credentials_for_hostname(hostname: str) -> bool:
    """Check if credentials are set for the given hostname."""
    return bool(_read_renku_token_for_hostname(hostname))


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def credentials(command: str, hostname: Optional[str]):
    """Credentials helper for git."""
    if command != "get":
        return

    # NOTE: hostname comes from the credential helper we set up and has proper format
    hostname = hostname or ""
    token = _read_renku_token_for_hostname(hostname=hostname) or ""

    communication.echo("username=renku")
    communication.echo(f"password={token}")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku ``status`` command."""

import os.path
import re
from datetime import datetime
from itertools import islice
from typing import TYPE_CHECKING, Dict, Generator, List, Optional, Tuple, cast

from renku.command.command_builder.command import Command
from renku.core.util import communication
from renku.domain_model.dataset import Dataset
from renku.domain_model.project_context import project_context
from renku.domain_model.provenance.activity import Activity
from renku.domain_model.workflow.plan import AbstractPlan

if TYPE_CHECKING:
    from renku.infrastructure.repository import Commit

CHECKPOINTS_PER_PAGE = 50


def rollback_command():
    """Command to perform a rollback fo the repo."""
    return Command().command(_rollback_command).require_clean().require_migration().with_database()


def _rollback_command():
    """Perform a rollback of the repo."""
    commits: Generator["Commit", None, None] = project_context.repository.iterate_commits(project_context.metadata_path)

    checkpoint = _prompt_for_checkpoint(commits)

    if not checkpoint:
        return

    diff = checkpoint[1].get_changes(commit="HEAD")

    confirmation_message, has_changes = _get_confirmation_message(diff)

    if not has_changes:
        communication.echo("There would be no changes rolling back to the selected command, exiting.")
        return

    communication.confirm(confirmation_message, abort=True)

    project_context.repository.reset(checkpoint[1], hard=True)


def _get_confirmation_message(diff) -> Tuple[str, bool]:
    """Create a confirmation message for changes that would be done by a rollback.

    Args:
        diff: Diff between two commits.

    Returns:
        Tuple[str, bool]: Tuple of confirmation message and if there would be changes.
    """
    modifications = _get_modifications_from_diff(diff)

    has_changes = False

    confirmation_message = "The following changes would be done:\n\nMetadata:\n"

    if modifications["metadata"]["restored"]:
        confirmation_message += "\nRestored ↻:\n\t" + "\n\t".join(modifications["metadata"]["restored"]) + "\n"
        has_changes = True

    if modifications["metadata"]["modified"]:
        confirmation_message += "\nModified ♻️:\n\t" + "\n\t".join(modifications["metadata"]["modified"]) + "\n"
        has_changes = True

    if modifications["metadata"]["removed"]:
        confirmation_message += "\nRemoved 🔥:\n\t" + "\n\t".join(modifications["metadata"]["removed"]) + "\n"
        has_changes = True

    confirmation_message += "\nFiles:\n"

    if modifications["files"]["restored"]:
        confirmation_message += "\nRestored ↻:\n\t" + "\n\t".join(modifications["files"]["restored"]) + "\n"
        has_changes = True

    if modifications["files"]["modified"]:
        confirmation_message += "\nModified ♻️:\n\t" + "\n\t".join(modifications["files"]["modified"]) + "\n"
        has_changes = True

    if modifications["files"]["removed"]:
        confirmation_message += "\nRemoved 🔥:\n\t" + "\n\t".join(modifications["files"]["removed"]) + "\n"
        has_changes = True

    confirmation_message += "\nProceed?"

    return confirmation_message, has_changes


def _get_modifications_from_diff(diff):
    """Get all modifications from a diff.

    Args:
        diff: Diff between two commits.

    Returns:
        List of metadata modifications made in diff.
    """
    modifications: Dict[str, Dict[str, List[str]]] = {
        "metadata": {"restored": [], "modified": [], "removed": []},
        "files": {"restored": [], "modified": [], "removed": []},
    }

    metadata_objects: Dict[str, Tuple[str, str, datetime]] = {}

    for diff_index in diff:
        entry = diff_index.a_path or diff_index.b_path
        entry_path = project_context.path / entry

        if str(project_context.database_path) == os.path.commonpath([project_context.database_path, entry_path]):
            # metadata file
            modification_type = _get_modification_type_from_db(entry)

            if not modification_type:
                continue

            entry, change_type, identifier, entry_date = modification_type

            if identifier not in metadata_objects or entry_date < metadata_objects[identifier][2]:
                # we only want he least recent change of a metadata object
                metadata_objects[identifier] = (entry, change_type, entry_date)

            continue

        elif str(project_context.metadata_path) == os.path.commonpath([project_context.metadata_path, entry_path]):
            # some other renku file
            continue

        # normal file
        if diff_index.added:
            modifications["files"]["removed"].append(entry)

        elif diff_index.deleted:
            modifications["files"]["restored"].append(entry)
        else:
            modifications["files"]["modified"].append(entry)

    for entry, change_type, _ in metadata_objects.values():
        modifications["metadata"][change_type].append(entry)

    return modifications


def _prompt_for_checkpoint(commits):
    """Ask to select a checkpoint to rollback to.

    Args:
        commits: Commits a user can choose from.

    Returns:
        Commit chosen by user.
    """
    checkpoint_iterator = _checkpoint_iterator(commits)

    all_checkpoints = []
    current_index = 0
    selected = None
    selection = None

    communication.echo("Select a checkpoint to roll back to:\n")

    # prompt user to select a checkpoint in batches
    while True:
        batch = list(islice(checkpoint_iterator, CHECKPOINTS_PER_PAGE))
        more_pages = len(batch) == CHECKPOINTS_PER_PAGE
        if batch:
            all_checkpoints.extend(batch)
            prompt = "\n".join(f"[{i}] {entry[0]}" for i, entry in enumerate(batch, current_index))
            prompt += "\nCheckpoint ([q] to quit"
            default = "q"
            if more_pages:
                prompt += ", [m] for more)"
                default = "m"
            else:
                prompt += ")"

            selection = communication.prompt(prompt, default=default)
        else:
            communication.echo("No more checkpoints.")

        while True:
            # loop until user makes a valid selection
            invalid = False
            if selection == "m" and more_pages:
                current_index += CHECKPOINTS_PER_PAGE
                break
            elif selection == "q":
                return
            elif selection is None:
                invalid = True
            else:
                try:
                    selected = int(selection)
                    if 0 <= selected < len(all_checkpoints):
                        break
                    else:
                        communication.warn("Not a valid checkpoint")
                        selected = None
                except (ValueError, TypeError):
                    invalid = True

            if invalid:
                communication.warn(
                    "Please enter a valid checkpoint number" + (", 'q' or 'm'" if more_pages else "or 'q'")
                )

            prompt = "Checkpoint ([q] to quit)"
            if more_pages:
                prompt += ", [m] for more)"
            else:
                prompt += ")"
            selection = communication.prompt("Checkpoint ([q] to quit)", default="q")

        if selected is not None:
            break

    if not all_checkpoints:
        communication.echo("No valid renku commands in project to roll back to.")
        return

    return all_checkpoints[selected]


def _get_modification_type_from_db(path: str) -> Optional[Tuple[str, str, str, datetime]]:
    """Get the modification type for an entry in the database.

    Args:
        path(str): Path to database object.

    Returns:
        Change information for object.
    """
    database = project_context.database
    db_object = database.get(os.path.basename(path))

    if isinstance(db_object, Activity):
        return (
            f"Run: {db_object.id} (Plan name: {db_object.association.plan.name})",
            "removed",
            db_object.id,
            db_object.ended_at_time,
        )
    elif isinstance(db_object, AbstractPlan):
        change_type = "removed"

        if db_object.derived_from:
            derived = database.get_by_id(db_object.derived_from)
            if db_object.name == derived.name:
                change_type = "modified"
        if db_object.date_removed:
            change_type = "restored"

        return (
            f"Plan: {db_object.name}",
            change_type,
            f"plan_{db_object.name}",
            db_object.date_removed or db_object.date_created,
        )
    elif isinstance(db_object, Dataset):
        change_type = "removed"

        if db_object.derived_from:
            change_type = "modified"
        if db_object.date_removed:
            change_type = "restored"

        return (
            f"Dataset: {db_object.name}",
            change_type,
            f"dataset_{db_object.name}",
            cast(datetime, db_object.date_removed or db_object.date_published or db_object.date_created),
        )
    else:
        return None


def _checkpoint_iterator(commits):
    """Iterate through commits to create checkpoints.

    Args:
        commits: Commits to iterate through.

    Returns:
        Iterator of commits that can be a checkpoint.
    """
    transaction_pattern = re.compile(r"\n\nrenku-transaction:\s([0-9a-g]+)$")

    current_checkpoint: Optional[Tuple[str, "Commit", str]] = None

    for commit in commits:
        commit_message = commit.message
        match = transaction_pattern.search(commit_message)

        if not match:
            continue

        transaction_id = match.group(0)
        entry = (
            f"{commit.authored_datetime:%Y-%m-%d %H:%M:%S} " f"\t{commit.hexsha[:7]}\t{commit_message.splitlines()[0]}",
            commit,
            transaction_id,
        )

        if not current_checkpoint:
            current_checkpoint = entry
            continue

        if transaction_id != current_checkpoint[2]:
            yield current_checkpoint
            current_checkpoint = entry

    if current_checkpoint:
        yield current_checkpoint



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku ``rerun`` command."""

from typing import List, Optional

from pydantic import validate_arguments

from renku.command.command_builder.command import Command, inject
from renku.core import errors
from renku.core.interface.activity_gateway import IActivityGateway
from renku.core.util.os import get_relative_paths
from renku.core.workflow.activity import get_activities_until_paths, sort_activities
from renku.core.workflow.execute import execute_workflow_graph
from renku.core.workflow.model.concrete_execution_graph import ExecutionGraph
from renku.domain_model.project_context import project_context


def rerun_command(skip_metadata_update: bool):
    """Recreate files generated by a sequence of ``run`` commands."""
    command = Command().command(_rerun).require_migration().require_clean()
    if skip_metadata_update:
        command = command.with_database(write=False)
    else:
        command = command.with_database(write=True).with_commit()
    return command


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def _rerun(
    dry_run: bool,
    sources: List[str],
    paths: List[str],
    provider: str,
    config: Optional[str],
    activity_gateway: IActivityGateway,
):
    """Rerun a previously run workflow.

    Args:
        dry_run (bool): Whether or not to actually execute the workflow or just show
            what would be executed.
        sources (List[str]): Input files to start execution from.
        paths (List[str]): Output paths to recreate.
        provider (str): Name of the workflow provider to use for execution.
        config (str): Path to configuration for the workflow provider.
        activity_gateway (IActivityGateway): Injected activity gateway.
    """

    sources = sources or []
    sources = get_relative_paths(base=project_context.path, paths=sources)
    paths = paths or []
    paths = get_relative_paths(base=project_context.path, paths=paths)

    activities = list(get_activities_until_paths(paths, sources, activity_gateway=activity_gateway))

    if len(activities) == 0:
        raise errors.NothingToExecuteError()

    activities = sort_activities(activities)
    if dry_run:
        return activities, set(sources)

    graph = ExecutionGraph([a.plan_with_values for a in activities], virtual_links=True)
    execute_workflow_graph(dag=graph.workflow_graph, provider=provider, config=config)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataset business logic."""

import imghdr
import os
import shutil
import urllib
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union, cast

import patoolib
from pydantic import validate_arguments

from renku.command.command_builder.command import inject
from renku.command.view_model.dataset import DatasetFileViewModel, DatasetViewModel
from renku.core import errors
from renku.core.config import get_value, remove_value, set_value
from renku.core.dataset.datasets_provenance import DatasetsProvenance
from renku.core.dataset.pointer_file import delete_external_file, is_linked_file_updated, update_linked_file
from renku.core.dataset.providers.api import AddProviderInterface, ProviderApi
from renku.core.dataset.providers.factory import ProviderFactory
from renku.core.dataset.providers.git import GitProvider
from renku.core.dataset.providers.models import DatasetUpdateAction, ProviderDataset
from renku.core.dataset.tag import get_dataset_by_tag, prompt_access_token, prompt_tag_selection
from renku.core.image import ImageObjectRequest
from renku.core.interface.dataset_gateway import IDatasetGateway
from renku.core.storage import check_external_storage, track_paths_in_storage
from renku.core.util import communication
from renku.core.util.datetime8601 import local_now
from renku.core.util.git import get_git_user
from renku.core.util.metadata import prompt_for_credentials, read_credentials, store_credentials
from renku.core.util.os import (
    create_symlink,
    delete_dataset_file,
    delete_path,
    get_absolute_path,
    get_file_size,
    get_files,
    get_relative_path,
    get_safe_relative_path,
    hash_file,
    is_path_empty,
    is_subpath,
    unmount_path,
)
from renku.core.util.tabulate import tabulate
from renku.core.util.urls import get_slug
from renku.core.util.util import parallel_execute
from renku.domain_model.constant import NO_VALUE, NON_EXISTING_ENTITY_CHECKSUM, NoValueType
from renku.domain_model.dataset import Dataset, DatasetDetailsJson, DatasetFile, RemoteEntity, is_dataset_slug_valid
from renku.domain_model.entity import Entity
from renku.domain_model.enums import ConfigFilter
from renku.domain_model.project_context import project_context
from renku.domain_model.provenance.agent import Person
from renku.domain_model.provenance.annotation import Annotation
from renku.infrastructure.immutable import DynamicProxy

if TYPE_CHECKING:
    from renku.core.interface.storage import IStorage


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def search_datasets(slug: str) -> List[str]:
    """Get all the datasets whose slug starts with the given string.

    Args:
        slug(str): Beginning of dataset slug to search for.

    Returns:
        List[str]: List of found dataset slugs.
    """
    datasets_provenance = DatasetsProvenance()
    return list(filter(lambda x: x.startswith(slug), map(lambda x: x.slug, datasets_provenance.datasets)))


def list_datasets():
    """List all datasets."""
    datasets_provenance = DatasetsProvenance()
    datasets = []

    for dataset in datasets_provenance.datasets:
        tags = datasets_provenance.get_all_tags(dataset)
        dataset = cast(Dataset, DynamicProxy(dataset))
        dataset.tags = tags
        dataset.tags_csv = ",".join(tag.name for tag in tags)
        dataset.datadir_path = str(dataset.get_datadir())
        datasets.append(dataset)

    return list(datasets)


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def create_dataset(
    slug: str,
    name: Optional[str] = None,
    description: Optional[str] = None,
    creators: Optional[List[Person]] = None,
    keywords: Optional[List[str]] = None,
    images: Optional[List[ImageObjectRequest]] = None,
    update_provenance: bool = True,
    custom_metadata: Optional[Dict[str, Any]] = None,
    storage: Optional[str] = None,
    datadir: Optional[Path] = None,
):
    """Create a dataset.

    Args:
        slug(str): Slug of the dataset
        name(Optional[str], optional): Dataset name (Default value = None).
        description(Optional[str], optional): Dataset description (Default value = None).
        creators(Optional[List[Person]], optional): Dataset creators (Default value = None).
        keywords(Optional[List[str]], optional): Dataset keywords (Default value = None).
        images(Optional[List[ImageObjectRequest]], optional): Dataset images (Default value = None).
        update_provenance(bool, optional): Whether to add this dataset to dataset provenance
            (Default value = True).
        custom_metadata(Optional[Dict[str, Any]], optional): Custom JSON-LD metadata (Default value = None).
        storage(Optional[str], optional): Backend storage's URI (Default value = None).
        datadir(Optional[Path]): Dataset's data directory (Default value = None).

    Returns:
        Dataset: The created dataset.
    """
    if not creators:
        creators = []
        user = get_git_user(repository=project_context.repository)

        if user:
            creators.append(user)

    if not is_dataset_slug_valid(slug):
        valid_slug = get_slug(slug, lowercase=False)
        raise errors.ParameterError(f"Dataset slug '{slug}' is not valid (Hint: '{valid_slug}' is valid).")

    datasets_provenance = DatasetsProvenance()

    if datasets_provenance.get_by_slug(slug=slug):
        raise errors.DatasetExistsError(slug)

    if not name:
        name = slug

    keywords = keywords or []

    annotations = (
        [Annotation(id=Annotation.generate_id(), source="renku", body=custom_metadata)] if custom_metadata else None
    )

    if datadir:
        try:
            datadir = get_safe_relative_path(datadir, project_context.path)
        except ValueError as e:
            raise errors.ParameterError("Datadir must be inside repository.") from e

    dataset = Dataset(
        identifier=None,
        slug=slug,
        name=name,
        description=description,
        creators=creators,
        keywords=keywords,
        project_id=project_context.project.id,
        annotations=annotations,
        storage=storage,
        datadir=datadir,
    )

    if images:
        set_dataset_images(dataset=dataset, images=images)

    if storage:
        provider = ProviderFactory.get_create_provider(uri=storage)
        provider.on_create(dataset=dataset)
    else:
        add_datadir_files_to_dataset(dataset)

    if update_provenance:
        datasets_provenance.add_or_update(dataset)

    return dataset


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def edit_dataset(
    slug: str,
    name: Optional[Union[str, NoValueType]],
    description: Optional[Union[str, NoValueType]],
    creators: Optional[Union[List[Person], NoValueType]],
    keywords: Optional[Union[List[str], NoValueType]] = NO_VALUE,
    images: Optional[Union[List[ImageObjectRequest], NoValueType]] = NO_VALUE,
    custom_metadata: Optional[Union[Dict, List[Dict], NoValueType]] = NO_VALUE,
    custom_metadata_source: Optional[Union[str, NoValueType]] = NO_VALUE,
):
    """Edit dataset metadata.

    Args:
        slug(str): Slug of the dataset to edit
        name(Optional[Union[str, NoValueType]]): New name for the dataset.
        description(Optional[Union[str, NoValueType]]): New description for the dataset.
        creators(Optional[Union[List[Person], NoValueType]]): New creators for the dataset.
        keywords(Optional[Union[List[str], NoValueType]]): New keywords for dataset (Default value = ``NO_VALUE``).
        images(Optional[Union[List[ImageObjectRequest], NoValueType]]): New images for dataset
            (Default value = ``NO_VALUE``).
        custom_metadata(Optional[Union[Dict, List[Dict], NoValueType]]): Custom JSON-LD metadata
            (Default value = ``NO_VALUE``).
        custom_metadata_source(Optional[Union[str, NoValueType]]): The custom metadata source
            (Default value = ``NO_VALUE``).

    Returns:
        bool: True if updates were performed.
    """
    if isinstance(name, str):
        name = name.strip()

    if name is None:
        name = ""

    possible_updates = {
        "creators": creators,
        "description": description,
        "keywords": keywords,
        "name": name,
    }

    dataset_provenance = DatasetsProvenance()
    dataset = dataset_provenance.get_by_slug(slug=slug)

    if dataset is None:
        raise errors.ParameterError("Dataset does not exist.")

    updated: Dict[str, Any] = {k: v for k, v in possible_updates.items() if v != NO_VALUE}

    if updated:
        dataset.update_metadata(creators=creators, description=description, keywords=keywords, name=name)

    if images == NO_VALUE:
        images_updated = False
    else:
        images_updated = set_dataset_images(dataset=dataset, images=cast(Optional[List[ImageObjectRequest]], images))

    if images_updated:
        updated["images"] = (
            None if images is None else [{"content_url": i.content_url, "position": i.position} for i in dataset.images]
        )

    if custom_metadata is not NO_VALUE:
        update_dataset_custom_metadata(
            dataset,
            cast(Optional[Union[Dict, List[Dict]]], custom_metadata),
            cast(Optional[str], custom_metadata_source),
        )
        updated["custom_metadata"] = custom_metadata

    if not updated:
        return []

    datasets_provenance = DatasetsProvenance()
    datasets_provenance.add_or_update(dataset, creator=get_git_user(project_context.repository))

    return updated


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def list_dataset_files(
    datasets: Optional[List[str]] = None,
    tag: Optional[str] = None,
    creators: Optional[Union[str, List[str], Tuple[str]]] = None,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
):
    """List dataset files.

    Args:
        datasets(Optional[List[str]]): Datasets to list files for (Default value = None).
        tag(str): Tag to filter by (Default value = None).
        creators(Optional[Union[str, List[str], Tuple[str]]]): Creators to filter by (Default value = None).
        include(Optional[List[str]]): Include filters for file paths (Default value = None).
        exclude(Optional[List[str]]): Exclude filters for file paths (Default value = None).

    Returns:
        List[DynamicProxy]: Filtered dataset files.
    """
    from renku.command.format.dataset_files import get_lfs_tracking_and_file_sizes

    records = filter_dataset_files(
        slugs=datasets, tag=tag, creators=creators, include=include, exclude=exclude, immutable=True
    )
    for record in records:
        record.title = record.dataset.name
        record.dataset_slug = record.dataset.slug
        record.dataset_id = record.dataset.id
        record.creators_csv = record.dataset.creators_csv
        record.creators_full_csv = record.dataset.creators_full_csv
        record.full_path = project_context.path / record.entity.path
        record.path = record.entity.path
        record.name = Path(record.entity.path).name
        record.added = record.date_added

    get_lfs_tracking_and_file_sizes(records, has_tag=bool(tag))

    return records


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def file_unlink(
    slug: str,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    yes: bool = False,
    dataset_files: Optional[List[DatasetFile]] = None,
):
    """Remove matching files from a dataset.

    Args:
        slug(str): Dataset slug.
        include(Optional[List[str]]): Include filter for files (Default value = None).
        exclude(Optional[List[str]]): Exclude filter for files (Default value = None).
        yes(bool): Whether to skip user confirmation or not (Default value = False).
        dataset_files(Optional[List[DatasetFile]]): Files to remove; ignore include and exclude if passed (Default value
            = None).

    Returns:
        List[DynamicProxy]: List of files that were removed.
    """
    repository = project_context.repository

    if not include and not exclude and not dataset_files:
        raise errors.ParameterError("include or exclude filters not specified.")

    datasets_provenance = DatasetsProvenance()

    dataset = datasets_provenance.get_by_slug(slug=slug)

    if not dataset:
        raise errors.ParameterError("Dataset does not exist.")

    records = []
    if not dataset_files:
        records = filter_dataset_files(slugs=[slug], include=include, exclude=exclude)
        if not records:
            raise errors.ParameterError("No records found.")
        dataset_files = [cast(DatasetFile, r) for r in records]

    if not yes:
        prompt_text = (
            f'You are about to remove following from "{slug}" dataset.'
            + "\n"
            + "\n".join([str(record.entity.path) for record in dataset_files])
            + "\nDo you wish to continue?"
        )
        communication.confirm(prompt_text, abort=True, warning=True)

    for file in dataset_files:
        dataset.unlink_file(file.entity.path)
        path_file = Path(file.entity.path)

        if file.is_external or file.linked:
            try:
                delete_external_file(file)
            except errors.InvalidFileOperation as e:
                communication.warn(f"Cannot delete dataset file {path_file}: {e}.")
        elif dataset.is_within_datadir(path_file):  # NOTE: Remove dataset file only if it's inside dataset's datadir
            datadir = dataset.get_datadir()

            if not path_file.exists():
                communication.warn(f"Dataset file {path_file} doesn't exist, skipping the removal from {datadir}.")
                continue

            try:
                if path_file.is_dir():
                    shutil.rmtree(str(path_file.absolute()), ignore_errors=False, onerror=None)
                else:
                    path_file.unlink()
            except Exception as err:
                communication.warn(f"Dataset file {path_file} could not be removed from {datadir} because of {err}.")

        repository.add(path_file)

    datasets_provenance.add_or_update(dataset, creator=get_git_user(repository))

    return records


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def remove_dataset(slug: str):
    """Delete a dataset.

    Args:
        slug(str): Slug of dataset to delete.
    """
    datasets_provenance = DatasetsProvenance()
    dataset = datasets_provenance.get_by_slug(slug=slug, strict=True)
    datasets_provenance.remove(dataset=dataset)


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def export_dataset(slug: str, provider_name: str, tag: Optional[str], **kwargs):
    """Export data to 3rd party provider.

    Args:
        slug(str): Slug of dataset to export.
        provider_name(str): Provider to use for export.
        tag(str): Dataset tag from which to export.
    """
    datasets_provenance = DatasetsProvenance()

    provider_name = provider_name.lower()

    # TODO: all these callbacks are ugly, improve in #737
    config_key_secret = "access_token"  # nosec

    dataset: Optional[Dataset] = datasets_provenance.get_by_slug(slug, strict=True, immutable=True)

    provider = ProviderFactory.get_export_provider(provider_name=provider_name)

    selected_tag = None
    tags = datasets_provenance.get_all_tags(dataset)  # type: ignore

    if tag:
        selected_tag = next((t for t in tags if t.name == tag), None)

        if not selected_tag:
            raise errors.ParameterError(f"Tag '{tag}' not found for dataset '{slug}'")
    elif tags:
        selected_tag = prompt_tag_selection(tags)

    if selected_tag:
        dataset = datasets_provenance.get_by_id(selected_tag.dataset_id.value, immutable=True)

        if not dataset:
            raise errors.DatasetNotFound(message=f"Cannot find dataset with id: '{selected_tag.dataset_id.value}'")

    dataset = cast(Dataset, DynamicProxy(dataset))

    exporter = provider.get_exporter(dataset=dataset, tag=selected_tag, **kwargs)

    if exporter.requires_access_token():
        access_token = read_credentials(section=provider_name, key=config_key_secret)

        if access_token is None:
            access_token = prompt_access_token(exporter)

            if access_token is None or len(access_token) == 0:
                raise errors.InvalidAccessToken()

            store_credentials(section=provider_name, key=config_key_secret, value=access_token)

        exporter.set_access_token(access_token)

    try:
        destination = exporter.export()
    except errors.AuthenticationError:
        remove_value(provider_name, config_key_secret, global_only=True)
        raise

    communication.echo(f"Exported to: {destination}")


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def import_dataset(
    uri: str,
    slug: Optional[str] = "",
    extract: bool = False,
    yes: bool = False,
    datadir: Optional[Path] = None,
    previous_dataset: Optional[Dataset] = None,
    delete: bool = False,
    gitlab_token: Optional[str] = None,
    **kwargs,
):
    """Import data from a 3rd party provider or another renku project.

    Args:
        uri(str): DOI or URL of dataset to import.
        slug(str): Slug to give to the imported dataset (Default value = "").
        extract(bool): Whether to extract compressed dataset data (Default value = False).
        yes(bool): Whether to skip user confirmation (Default value = False).
        datadir(Optional[Path]): Dataset's data directory (Default value = None).
        previous_dataset(Optional[Dataset]): Previously imported dataset version (Default value = None).
        delete(bool): Whether to delete files that don't exist anymore (Default value = False).
        gitlab_token(Optional[str]): Gitlab OAuth2 token (Default value = None).
    """
    from renku.core.dataset.dataset_add import add_to_dataset

    def confirm_download(files):
        if yes:
            return

        headers = {"checksum": "checksum", "filename": "name", "filesize_str": "size", "filetype": "type"}
        communication.echo(tabulate(files, headers=headers, floatfmt=".2f"))
        communication.confirm("Do you wish to download this version?", abort=True, warning=True)

    def calculate_total_size(files) -> int:
        total_size = 0
        for file in files:
            if file.filesize is not None:
                total_size += file.filesize

        return total_size

    def remove_files(dataset):
        """Remove files that exist in ``previous_dataset`` but not in ``dataset``.

        Args:
            dataset(Dataset): Dataset to update.
        """
        if not delete or not previous_dataset:
            return

        current_paths = {str(f.entity.path) for f in dataset.files}
        previous_paths = {str(f.entity.path) for f in previous_dataset.files}
        deleted_paths = previous_paths - current_paths

        for path in deleted_paths:
            delete_dataset_file(project_context.path / path, follow_symlinks=True)

    provider = ProviderFactory.get_import_provider(uri)

    try:
        importer = provider.get_importer(gitlab_token=gitlab_token, **kwargs)
        provider_dataset: ProviderDataset = importer.fetch_provider_dataset()
    except KeyError as e:
        raise errors.ParameterError(f"Could not process '{uri}'.\nUnable to fetch metadata: {e}")
    except LookupError as e:
        raise errors.ParameterError(f"Could not process '{uri}'.\nReason: {e}")

    if not importer.provider_dataset_files:
        raise errors.ParameterError(f"Dataset '{uri}' has no files.")

    confirm_download(importer.provider_dataset_files)

    try:
        if not importer.is_latest_version():
            communication.warn(f"Newer version found at {importer.latest_uri}")
    except (KeyError, LookupError):
        pass

    if datadir and previous_dataset:
        raise errors.ParameterError("Can't specify datadir when updating a previously imported dataset.")
    elif datadir:
        try:
            datadir = get_safe_relative_path(datadir, project_context.path)
        except ValueError as e:
            raise errors.ParameterError("Datadir must be inside repository.") from e

    slug = slug or provider_dataset.slug

    new_dataset = add_to_dataset(
        dataset_slug=slug,
        urls=[],
        importer=importer,
        create=not previous_dataset,
        force=True,  # NOTE: Force-add to include any ignored files
        extract=extract,
        overwrite=True,
        total_size=calculate_total_size(importer.provider_dataset_files),
        clear_files_before=True,
        datadir=datadir,
        storage=provider_dataset.storage,
    )

    new_dataset.update_metadata_from(provider_dataset)
    # NOTE: Remove derived_from because this is an updated and imported dataset and won't be a derivative
    new_dataset.derived_from = None

    remove_files(new_dataset)

    importer.tag_dataset(slug)
    importer.copy_extra_metadata(new_dataset)

    project_context.database.commit()


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def update_datasets(
    slugs: List[str],
    creators: Optional[str],
    include: Optional[List[str]],
    exclude: Optional[List[str]],
    ref: Optional[str],
    delete: bool,
    no_local: bool,
    no_remote: bool,
    check_data_directory: bool,
    update_all: bool,
    dry_run: bool,
    plain: bool,
    dataset_gateway: IDatasetGateway,
) -> Tuple[List[DatasetViewModel], List[DatasetFileViewModel]]:
    """Update dataset files.

    Args:
        slugs(List[str]): Slugs of datasets to update.
        creators(Optional[str]): Creators to filter dataset files by.
        include(Optional[List[str]]): Include filter for paths to update.
        exclude(Optional[List[str]]): Exclude filter for paths to update.
        ref(Optional[str]): Git reference to use for update.
        delete(bool): Whether to delete files that don't exist on remote anymore.
        no_local(bool): Whether to exclude local files from the update.
        no_remote(bool): Whether to exclude remote files from the update.
        check_data_directory(bool): Whether to check the dataset's data directory for new files.
        update_all(bool): Whether to update all datasets.
        dry_run(bool): Whether to return a preview of what would be updated.
        plain(bool): Whether plain output should be produced.
        dataset_gateway(IDatasetGateway): Injected dataset gateway.
    """
    from renku.core.dataset.providers.renku import RenkuProvider

    if not update_all and not slugs and not include and not exclude and not dry_run:
        raise errors.ParameterError("No update criteria is specified")

    imported_dataset_updates: List[Dataset] = []

    all_datasets = dataset_gateway.get_all_active_datasets()
    imported_datasets = [d for d in all_datasets if d.same_as]

    if slugs and update_all:
        raise errors.ParameterError("Cannot pass dataset slugs when updating all datasets")
    elif (include or exclude) and update_all:
        raise errors.ParameterError("Cannot specify include and exclude filters when updating all datasets")
    elif (include or exclude) and slugs and any(d for d in imported_datasets if d.slug in slugs):
        raise errors.IncompatibleParametersError(first_param="--include/--exclude", second_param="imported datasets")

    slugs = slugs or [d.slug for d in all_datasets]

    # NOTE: update imported datasets
    if not include and not exclude and not no_remote:
        must_match_records = False

        for dataset in imported_datasets:
            if dataset.slug not in slugs:
                continue

            uri = dataset.same_as.value  # type: ignore
            try:
                provider = ProviderFactory.get_import_provider(uri)
            except errors.DatasetProviderNotFound:
                continue

            record = provider.get_importer()

            if isinstance(provider, RenkuProvider) and dataset.version is not None:
                tags = dataset_gateway.get_all_tags(dataset=dataset)
                tag = next((t for t in tags if t.name == dataset.version), None)
                # NOTE: Do not update Renku dataset that are imported from a specific version
                if tag is not None and tag.dataset_id.value == dataset.id:
                    communication.echo(
                        f"Skipped updating imported Renku dataset '{dataset.slug}' with tag '{tag.name}'"
                    )
                    slugs.remove(dataset.slug)
                    continue

            if record.is_latest_version() and record.is_version_equal_to(dataset):
                slugs.remove(dataset.slug)
                continue

            if not dry_run:
                uri = record.latest_uri

                # NOTE: set extract to false if there are any archives present in the dataset
                extract = True
                for f in dataset.files:
                    try:
                        patoolib.get_archive_format(f.entity.path)
                    except patoolib.util.PatoolError:
                        continue
                    else:
                        extract = False
                        break

                import_dataset(
                    uri=uri, slug=dataset.slug, extract=extract, yes=True, previous_dataset=dataset, delete=delete
                )

                communication.echo(f"Updated dataset '{dataset.slug}' from remote provider")

            slugs.remove(dataset.slug)
            imported_dataset_updates.append(dataset)
    else:
        must_match_records = True

    imported_dataset_updates_view_models = [DatasetViewModel.from_dataset(d) for d in imported_dataset_updates]

    if not slugs:
        return imported_dataset_updates_view_models, []

    # NOTE: Exclude all imported dataset from individual file filter
    records = filter_dataset_files(
        slugs=slugs,
        creators=creators,
        include=include,
        exclude=exclude,
        ignore=[d.slug for d in imported_datasets],
        check_data_directory=check_data_directory,
    )

    if not records:
        if must_match_records and not plain:
            raise errors.ParameterError("No files matched the criteria.")
        return imported_dataset_updates_view_models, []

    provider_files: Dict[AddProviderInterface, List[DynamicProxy]] = defaultdict(list)
    unique_remotes = set()
    linked_files = []

    for file in records:
        if file.linked:
            linked_files.append(file)
        else:
            if not getattr(file, "provider", None):
                if file.based_on:
                    uri = file.dataset.same_as.value if file.dataset.same_as else file.based_on.url
                else:
                    uri = file.source
                try:
                    file.provider = cast(
                        AddProviderInterface,
                        ProviderFactory.get_add_provider(uri),
                    )
                except errors.DatasetProviderNotFound:
                    communication.warn(f"Couldn't find provider for file {file.path} in dataset {file.dataset.slug}")
                    continue

            provider_files[file.provider].append(file)

            if isinstance(file.provider, GitProvider):
                unique_remotes.add(file.based_on.url)

    if ref and len(unique_remotes) > 1:
        raise errors.ParameterError(
            "Cannot specify a reference with more than one Git repository.\n"
            "Limit list of files to be updated to one repository. See 'renku dataset update -h' for more information."
        )

    updated_files: List[DynamicProxy] = []
    deleted_files: List[DynamicProxy] = []

    if linked_files:
        updated = update_linked_files(linked_files, dry_run=dry_run)
        updated_files.extend(updated)

    provider_context: Dict[str, Any] = {}

    for provider, files in provider_files.items():
        if (no_remote and cast(ProviderApi, provider).is_remote) or (
            no_local and not cast(ProviderApi, provider).is_remote
        ):
            continue

        results = provider.update_files(
            files=files,
            dry_run=dry_run,
            delete=delete,
            context=provider_context,
            ref=ref,
            check_data_directory=check_data_directory,
        )
        updated_files.extend(r.entity for r in results if r.action == DatasetUpdateAction.UPDATE)
        deleted_files.extend(r.entity for r in results if r.action == DatasetUpdateAction.DELETE)

    if not dry_run:
        if deleted_files and not delete:
            communication.echo("Some files are deleted: Pass '--delete' to remove them from datasets' metadata")
        if updated_files or (deleted_files and delete):
            file_paths = {str(project_context.path / f.entity.path) for f in updated_files}
            # Force-add to include possible ignored files that are in datasets
            repository = project_context.repository
            repository.add(*file_paths, force=True)
            repository.add(project_context.pointers_path, force=True)

            _update_datasets_files_metadata(updated_files, deleted_files, delete)

        message = f"Updated {len(updated_files)} files"
        if delete:
            message += f" and deleted {len(deleted_files)} files"
        communication.echo(message)
    else:
        for file in deleted_files:
            if not file.date_removed:
                file.date_removed = local_now()

    dataset_files_view_models = [
        DatasetFileViewModel.from_dataset_file(cast(DatasetFile, f), f.dataset) for f in updated_files + deleted_files
    ]
    return imported_dataset_updates_view_models, dataset_files_view_models


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def show_dataset(slug: str, tag: Optional[str] = None):
    """Show detailed dataset information.

    Args:
        slug(str): Slug of dataset to show details for.
        tag(str, optional): Tags for which to get the metadata (Default value = None).

    Returns:
        dict: JSON dictionary of dataset details.
    """
    datasets_provenance = DatasetsProvenance()
    dataset: Optional[Dataset] = datasets_provenance.get_by_slug(slug, strict=True)

    if tag is None:
        return DatasetDetailsJson().dump(dataset)

    tags = datasets_provenance.get_all_tags(dataset=cast(Dataset, dataset))

    selected_tag = next((t for t in tags if t.name == tag), None)

    if selected_tag is None:
        raise errors.DatasetTagNotFound(tag)

    dataset = datasets_provenance.get_by_id(selected_tag.dataset_id.value)
    return DatasetDetailsJson().dump(dataset)


def add_datadir_files_to_dataset(dataset: Dataset) -> None:
    """Add all files in a datasets data directory to the dataset.

    Args:
        dataset(Dataset): The dataset to add data dir files to.
    """
    datadir = get_safe_relative_path(dataset.get_datadir(), project_context.path)

    if datadir.exists():
        # NOTE: Add existing files to dataset
        dataset_files: List[DatasetFile] = []
        files: List[Path] = []
        existing_files: List[Union[Path, str]] = list(get_files(datadir))
        checksums = project_context.repository.get_object_hashes(existing_files)

        for file in cast(List[Path], existing_files):
            files.append(file)
            dataset_files.append(DatasetFile.from_path(path=file, source=file, checksum=checksums.get(file)))

        if not dataset_files:
            return

        if check_external_storage():
            track_paths_in_storage(*files)
        project_context.repository.add(*files)

        dataset.add_or_update_files(dataset_files)


def set_dataset_images(dataset: Dataset, images: Optional[List[ImageObjectRequest]]):
    """Set a dataset's images.

    Args:
        dataset(Dataset): The dataset to set images on.
        images(List[ImageObjectRequest]): The images to set.

    Returns:
        True if images were set/modified.
    """
    if not images:
        images = []

    image_folder = project_context.dataset_images_path / dataset.initial_identifier
    image_folder.mkdir(exist_ok=True, parents=True)
    previous_images = dataset.images or []

    dataset.images = []
    images_updated = False
    for img in images:
        image_folder = project_context.dataset_images_path / dataset.initial_identifier
        try:
            img_object = img.to_image_object(owner_id=dataset.id)
        except errors.ImageError as e:
            raise errors.DatasetImageError(e) from e

        path = img_object.content_url

        if not img_object.is_remote:
            # NOTE: only copy dataset image if it's not in .renku/datasets/<id>/images/ already
            if not path.startswith(str(image_folder)):
                image_type = imghdr.what(path)
                if image_type:
                    ext = f".{image_type}"
                else:
                    _, ext = os.path.splitext(path)
                target_image_path: Union[Path, str] = image_folder / f"{img_object.position}{ext}"

                image_folder.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy(path, target_image_path)
            else:
                target_image_path = path

            img_object.content_url = get_relative_path(target_image_path, base=project_context.path)  # type: ignore

        if any(i.position == img_object.position for i in dataset.images):
            raise errors.DatasetImageError(f"Duplicate dataset image specified for position {img_object.position}")

        dataset.images.append(img_object)
        images_updated = True

    new_urls = [i.content_url for i in dataset.images]

    for prev in previous_images:
        # NOTE: Delete images if they were removed
        if prev.content_url in new_urls or urllib.parse.urlparse(prev.content_url).netloc:
            continue

        path = prev.content_url
        if not os.path.isabs(path):
            path = os.path.normpath(os.path.join(project_context.path, path))

        os.remove(path)

    return images_updated or dataset.images != previous_images


def update_dataset_custom_metadata(
    dataset: Dataset,
    custom_metadata: Optional[Union[Dict, List[Dict]]],
    custom_metadata_source: Optional[str],
):
    """Update custom metadata on a dataset.

    Args:
        dataset(Dataset): The dataset to update.
        custom_metadata(Dict): Custom JSON-LD metadata to set.
        custom_metadata_source(str): The source field for the custom metadata.
    """

    existing_metadata = [a for a in dataset.annotations if a.source != custom_metadata_source]

    if custom_metadata is not None and custom_metadata_source is not None:
        if isinstance(custom_metadata, dict):
            custom_metadata = [custom_metadata]
        for cm in custom_metadata:
            existing_metadata.append(Annotation(id=Annotation.generate_id(), body=cm, source=custom_metadata_source))

    dataset.annotations = existing_metadata


@inject.autoparams("dataset_gateway")
def move_files(dataset_gateway: IDatasetGateway, files: Dict[Path, Path], to_dataset_slug: Optional[str] = None):
    """Move files and their metadata from one or more datasets to a target dataset.

    Args:
        dataset_gateway(IDatasetGateway):Injected dataset gateway.
        files(Dict[Path, Path]): Files to move
        to_dataset_slug(Optional[str], optional): Target dataset (Default value = None)
    """
    datasets = [d.copy() for d in dataset_gateway.get_all_active_datasets()]

    to_dataset: Optional[Dataset] = None
    if to_dataset_slug:
        # NOTE: Use the same dataset object or otherwise a race happens if dataset is in both source and destination
        to_dataset = next(d for d in datasets if d.slug == to_dataset_slug)
    modified_datasets: Dict[str, Dataset] = {}

    progress_name = "Updating dataset metadata"
    communication.start_progress(progress_name, total=len(files))
    try:
        checksums = project_context.repository.get_object_hashes(
            [file.relative_to(project_context.path) for file in files.values()]
        )
        for src, dst in files.items():
            src = src.relative_to(project_context.path)
            dst = dst.relative_to(project_context.path)
            # NOTE: Files are moved at this point, so, we can use dst
            new_dataset_file = DatasetFile.from_path(dst, checksum=checksums.get(dst))

            for dataset in datasets:
                removed = dataset.unlink_file(src, missing_ok=True)
                if removed:
                    modified_datasets[dataset.slug] = dataset
                    new_dataset_file.based_on = removed.based_on
                    new_dataset_file.source = removed.source

                    if not to_dataset and (
                        new_dataset_file.linked
                        or is_subpath(project_context.path / dst, project_context.path / dataset.get_datadir())
                    ):
                        dataset.add_or_update_files(new_dataset_file)

                # NOTE: Update dataset if it contains a destination that is being overwritten
                modified = dataset.find_file(dst)
                added = is_subpath(project_context.path / dst, project_context.path / dataset.get_datadir())
                if modified or added:
                    modified_datasets[dataset.slug] = dataset
                    dataset.add_or_update_files(new_dataset_file)

            if to_dataset:
                to_dataset.add_or_update_files(new_dataset_file)

            communication.update_progress(progress_name, amount=1)
    finally:
        communication.finalize_progress(progress_name)

    datasets_provenance = DatasetsProvenance()
    modified_dataset_values = list(modified_datasets.values())
    creator = get_git_user(repository=project_context.repository)
    for modified_dataset in modified_dataset_values:
        datasets_provenance.add_or_update(modified_dataset, creator=creator)
    if to_dataset and to_dataset not in modified_dataset_values:
        datasets_provenance.add_or_update(to_dataset, creator=creator)


def _update_datasets_files_metadata(updated_files: List[DynamicProxy], deleted_files: List[DynamicProxy], delete: bool):
    modified_datasets = {}
    checksums = project_context.repository.get_object_hashes([file.entity.path for file in updated_files])
    for file in updated_files:
        new_file = DatasetFile.from_path(
            path=file.entity.path, based_on=file.based_on, source=file.source, checksum=checksums.get(file.entity.path)
        )
        modified_datasets[file.dataset.slug] = (
            file.dataset._subject if isinstance(file.dataset, DynamicProxy) else file.dataset
        )
        file.dataset.add_or_update_files(new_file)

    if delete:
        for file in deleted_files:
            modified_datasets[file.dataset.slug] = (
                file.dataset._subject if isinstance(file.dataset, DynamicProxy) else file.dataset
            )
            file.dataset.unlink_file(file.entity.path)

    datasets_provenance = DatasetsProvenance()
    for dataset in modified_datasets.values():
        datasets_provenance.add_or_update(dataset, creator=get_git_user(repository=project_context.repository))


def update_linked_files(records: List[DynamicProxy], dry_run: bool) -> List[DynamicProxy]:
    """Update files linked to other files in the project.

    Args:
        records(List[DynamicProxy]): File records to update.
        dry_run(bool): Whether to return a preview of what would be updated.
    """
    updated_files = []

    for file in records:
        if file.linked:
            try:
                updated, checksum = is_linked_file_updated(path=file.entity.path)
            except errors.ExternalFileNotFound as e:
                if not dry_run:
                    raise
                communication.warn(str(e))
                continue

            if updated:
                if not dry_run:
                    update_linked_file(path=file.entity.path, checksum=checksum)
                updated_files.append(file)

    return updated_files


@inject.autoparams("dataset_gateway")
def filter_dataset_files(
    dataset_gateway: IDatasetGateway,
    slugs: Optional[List[str]] = None,
    tag: Optional[str] = None,
    creators: Optional[Union[str, List[str], Tuple[str]]] = None,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    ignore: Optional[List[str]] = None,
    immutable: bool = False,
    check_data_directory: bool = False,
) -> List[DynamicProxy]:
    """Filter dataset files by specified filters.

    Args:
        dataset_gateway(IDatasetGateway):Injected dataset gateway.
        slugs(Optional[List[str]]): Filter by specified dataset slugs (Default value = None).
        tag(Optional[str]): Filter by specified tag (Default value = None).
        creators(Optional[Union[str, List[str], Tuple[str]]]): Filter by creators (Default value = None).
        include(Optional[List[str]]): Tuple containing patterns to which include from result (Default value = None).
        exclude(Optional[List[str]]): Tuple containing patterns to which exclude from result (Default value = None).
        ignore(Optional[List[str]]): Ignored datasets (Default value = None).
        immutable(bool): Return immutable copies of dataset objects (Default value = False).
        check_data_directory(bool): Whether to check for new files in dataset's data directory that aren't in the
            dataset yet (Default value = False).
    Returns:
        List[DynamicProxy]: List of filtered files sorted by date added.
    """

    def should_include(filepath: Path) -> bool:
        """Check if file matches one of include filters and not in exclude filter."""
        if exclude:
            for pattern in exclude:
                if filepath.match(pattern):
                    return False

        if include:
            for pattern in include:
                if filepath.match(pattern):
                    return True
            return False

        return True

    creators_set = set()
    if isinstance(creators, str):
        creators_set = set(creators.split(","))
    elif isinstance(creators, list) or isinstance(creators, tuple):
        creators_set = set(creators)

    records = []
    unused_slugs = set(slugs) if slugs is not None else set()

    if ignore:
        unused_slugs = unused_slugs - set(ignore)

    for dataset in dataset_gateway.get_all_active_datasets():
        if (slugs and dataset.slug not in slugs) or (ignore and dataset.slug in ignore):
            continue

        if tag:
            dataset = get_dataset_by_tag(dataset=dataset, tag=tag)  # type: ignore
            if not dataset:
                continue

        if not immutable:
            dataset = dataset.copy()

        if unused_slugs:
            unused_slugs.remove(dataset.slug)

        if creators_set:
            dataset_creators = {creator.name for creator in dataset.creators}
            if not creators_set.issubset(dataset_creators):
                continue

        for file in dataset.files:
            if not should_include(Path(file.entity.path)):
                continue

            record = DynamicProxy(file)
            record.dataset = DynamicProxy(dataset)
            records.append(record)

        if not check_data_directory:
            continue

        for root, _, files in os.walk(project_context.path / dataset.get_datadir()):
            current_folder = Path(root)
            for current_file in files:
                file_path = get_safe_relative_path(current_folder / current_file, project_context.path)
                if should_include(file_path) and not dataset.find_file(file_path):
                    # New file in dataset folder
                    record = DynamicProxy(DatasetFile.from_path(file_path))
                    record.dataset = dataset
                    records.append(record)

    if unused_slugs:
        unused_slugs_str = ", ".join(unused_slugs)
        raise errors.ParameterError(f"These datasets don't exist: {unused_slugs_str}")

    return sorted(records, key=lambda r: r.date_added)


def download_file(file: DatasetFile, storage: "IStorage") -> List[DatasetFile]:
    """Download a dataset file and retrieve its missing metadata (if any).

    Args:
        file(DatasetFile): Dataset file to download.
        storage: Dataset's cloud storage (an instance of ``IStorage``).

    Returns:
         List[DatasetFile]: A list with the updated file if its metadata was missing; an empty list otherwise.

    """
    if not file.based_on:
        raise errors.DatasetImportError(f"Dataset file doesn't have a URI: {file.entity.path}")

    path = project_context.path / file.entity.path
    path.resolve().parent.mkdir(parents=True, exist_ok=True)

    # NOTE: Don't check if destination file exists. ``IStorage.copy`` won't copy a file if it exists and is not
    # modified.

    communication.start_progress(name=file.entity.path, total=1)
    try:
        storage.download(file.based_on.url, path)
        communication.update_progress(name=file.entity.path, amount=1)
    finally:
        communication.finalize_progress(name=file.entity.path)

    # NOTE: File has no missing information
    if file.has_valid_checksum() and file.has_valid_size():
        return []

    if not file.has_valid_checksum():
        md5_hash = hash_file(path, hash_type="md5") or NON_EXISTING_ENTITY_CHECKSUM
        entity = Entity(path=file.entity.path, checksum=md5_hash)
        remote_entity = RemoteEntity(checksum=md5_hash, url=file.based_on.url, path=file.based_on.path)
    else:
        entity = file.entity
        remote_entity = file.based_on

    size = file.size if file.has_valid_size() else get_file_size(path)

    return [
        DatasetFile(
            entity=entity,
            based_on=remote_entity,
            size=size,
            date_added=file.date_added,
            date_removed=file.date_removed,
            source=file.source,
        )
    ]


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def pull_cloud_storage(slug: str, location: Optional[Path] = None) -> None:
    """Pull/copy data for a cloud storage to a dataset's data directory or a specified location.

    Args:
        slug(str): Slug of the dataset
        location(Optional[Path]): A directory to copy data to (Default value = None).
    """
    dataset, datadir = _get_dataset_with_cloud_storage(slug=slug)

    # NOTE: Try to unmount the path in case it was mounted before
    unmount_path(datadir)

    if location:
        if not is_path_empty(datadir):
            communication.confirm(
                f"Dataset's data directory will be removed: {dataset.get_datadir()}. Do you want to continue?",
                abort=True,
                warning=True,
            )
        create_symlink(target=location, symlink_path=datadir, overwrite=True)

    provider = ProviderFactory.get_pull_provider(uri=dataset.storage)
    storage = provider.get_storage()

    updated_files = parallel_execute(download_file, dataset.files, rate=5, storage=storage)

    if updated_files:
        dataset.add_or_update_files(updated_files)
        DatasetsProvenance().add_or_update(dataset, creator=get_git_user(repository=project_context.repository))
        project_context.database.commit()


def store_dataset_data_location(dataset: Dataset, location: Optional[Path]) -> None:
    """Store data location for a dataset in the config file."""
    section = "dataset-locations"
    key = dataset.slug

    if not location:
        remove_value(section=section, key=key)
    else:
        set_value(section=section, key=key, value=get_absolute_path(location))


def read_dataset_data_location(dataset: Dataset) -> Optional[str]:
    """Read data location for a dataset in the config file."""
    return get_value(section="dataset-locations", key=dataset.slug, config_filter=ConfigFilter.LOCAL_ONLY)


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def mount_cloud_storage(slug: str, existing: Optional[Path], yes: bool) -> None:
    """Mount a cloud storage to a dataset's data directory.

    Args:
        slug(str): Slug of the dataset
        existing(Optional[Path]): An existing mount point to use instead of actually mounting the backend storage.
        yes(bool): Don't prompt when removing non-empty dataset's data directory.
    """
    dataset, datadir = _get_dataset_with_cloud_storage(slug=slug)

    # NOTE: Try to unmount the path in case it was mounted before
    unmount_path(datadir)

    if not is_path_empty(datadir) and not yes:
        communication.confirm(
            f"Dataset's data directory will be removed: {dataset.get_datadir()}. Do you want to continue?",
            abort=True,
            warning=True,
        )

    if existing:
        create_symlink(target=existing, symlink_path=datadir, overwrite=True)
        return

    delete_path(datadir)
    datadir.mkdir(parents=True, exist_ok=True)

    provider = ProviderFactory.get_mount_provider(uri=dataset.storage)
    credentials = provider.get_credentials()
    prompt_for_credentials(credentials)
    storage = provider.get_storage(credentials=credentials)

    with communication.busy(f"Mounting {provider.uri}"):
        storage.mount(datadir)


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def unmount_cloud_storage(slug: str) -> None:
    """Mount a cloud storage to a dataset's data directory.

    Args:
        slug(str): Slug of the dataset
    """
    _, datadir = _get_dataset_with_cloud_storage(slug=slug)
    unmount_path(datadir)


def _get_dataset_with_cloud_storage(slug: str) -> Tuple[Dataset, Path]:
    datasets_provenance = DatasetsProvenance()

    dataset = datasets_provenance.get_by_slug(slug=slug, strict=True)

    if not dataset.storage:
        raise errors.ParameterError(f"Dataset '{slug}' doesn't have a storage backend")

    datadir = project_context.path / dataset.get_datadir()

    return dataset, datadir



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for renku service controllers."""
import signal

from renku.ui.service.errors import IntermittentTimeoutError


def timeout(fn, fn_args=None, fn_kwargs=None, timeout_duration=3600, default=None):
    """Execute provided function or timeout."""
    fn_args = fn_args or ()
    fn_kwargs = fn_kwargs or {}

    def signal_handler(signum, frame):
        raise IntermittentTimeoutError()

    # NOTE: Set the timeout handler.
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(timeout_duration)

    try:
        result = fn(*fn_args, **fn_kwargs)
    except IntermittentTimeoutError:
        result = default
    finally:
        signal.alarm(0)

    return result



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache project related serializers."""
import uuid
from datetime import datetime

from marshmallow import ValidationError, fields, post_load, validates_schema

from renku.ui.service.cache.models.project import Project
from renku.ui.service.serializers.common import AccessSchema, CreationSchema, MandatoryUserSchema
from renku.ui.service.utils import normalize_git_url


class ProjectSchema(CreationSchema, AccessSchema, MandatoryUserSchema):
    """Context schema for project clone."""

    last_fetched_at = fields.DateTime(load_default=datetime.utcnow)

    project_id = fields.String(load_default=lambda: uuid.uuid4().hex)

    clone_depth = fields.Integer()
    git_url = fields.String()

    name = fields.String(required=True)
    slug = fields.String(required=True)
    description = fields.String(load_default=None)
    owner = fields.String(required=True)
    initialized = fields.Boolean(dump_default=False)
    commit_sha = fields.String(required=False, load_default=None, dump_default=None)
    branch = fields.String(required=False, load_default=None, dump_default=None)

    @post_load
    def make_project(self, data, **options):
        """Construct project object."""
        if data.get("git_url"):
            data["git_url"] = normalize_git_url(data["git_url"])
        data["name"] = normalize_git_url(data["name"])
        data["slug"] = normalize_git_url(data["slug"])
        return Project(**data)

    @validates_schema
    def ensure_only_commit_sha_or_branch(self, data, **kwargs):
        """Checks that only a commit SHA or branch is set and not both."""
        if data.get("commit_sha") and data.get("branch"):
            raise ValidationError("You cannot specify a commit SHA and a branch, only one or the other")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""JSON-LD SHACL validations."""

from typing import Union

from pyshacl import validate

try:
    import importlib_resources  # type: ignore[import]
except ImportError:
    import importlib.resources as importlib_resources  # type: ignore


def validate_graph(graph, shacl_path=None, format="nquads"):
    """Validate the current graph with a SHACL schema.

    Uses default schema if not supplied.
    """
    shacl: Union[str, bytes]
    if shacl_path:
        with open(shacl_path, encoding="utf-8") as f:
            shacl = f.read()
    else:
        shacl = importlib_resources.files("renku.data").joinpath("shacl_shape.json").read_bytes()

    return validate(
        graph,
        shacl_graph=shacl,
        inference="rdfs",
        meta_shacl=True,
        debug=False,
        data_graph_format=format,
        shacl_graph_format="json-ld",
        advanced=True,
    )
